#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level **simulation phase cache** (e.g., container persisting previously
constructed large-scale objects for a simulation phase) functionality.
'''

# ....................{ IMPORTS                            }....................
from abc import ABCMeta  # , abstractmethod
from betse.science.phase.phasecls import SimPhase
from betse.util.py import pyref
from betse.util.type.types import type_check

# ....................{ CLASSES                            }....................
class SimPhaseCaches(object):
    '''
    Namespace containing all simulation phase-specific caches, each persisting
    previously constructed large-scale objects for subsequent reuse by a single
    simulation phase.

    Attributes
    ----------
    upscaled : SimPhaseCacheCellsUpscaled
        Subcache of all upscaled objects constructed for this phase.
    vector : SimPhaseCacheVectorCells
        Subcache of all vectors constructed for this phase.
    vector_field : SimPhaseCacheVectorFieldCells
        Subcache of all vector fields constructed for this phase.
    '''

    # ..................{ INITIALIZORS                       }..................
    @type_check
    def __init__(self, phase: SimPhase) -> None:
        '''
        Initialize this simulation phase cache.

        Parameters
        ----------
        phase : SimPhase
            Parent simulation phase.
        '''

        # Avoid circular import dependencies.
        from betse.science.math.cache.cacheupscaled import (
            SimPhaseCacheUpscaled)
        from betse.science.math.cache.cachevec import SimPhaseCacheVectorCells
        from betse.science.math.cache.cachevecfld import (
            SimPhaseCacheVectorFieldCells)

        # Classify all subcaches imported above.
        self.upscaled = SimPhaseCacheUpscaled(phase)
        self.vector = SimPhaseCacheVectorCells(phase)
        self.vector_field = SimPhaseCacheVectorFieldCells(phase)

# ....................{ SUPERCLASSES                       }....................
class SimPhaseCacheABC(object, metaclass=ABCMeta):
    '''
    Abstract base class of all **simulation phase subcache** (i.e., container
    persisting previously constructed large-scale objects for some facet of a
    simulation phase) subclasses.

    Design
    ----------
    Subcaches provide namespace isolation but are otherwise purely superficial.
    Technically, all objects cached by a subcache could simply be cached by the
    parent :class:`SimPhaseCaches` object instead. Doing so would quickly become
    cumbersome, however, both for maintenance and reuse.

    Attributes
    ----------
    _phase : SimPhase
        Parent simulation phase. This attribute is technically accessible as
        :attr:`_cache.phase` but is provided by this class for convenience.
    '''

    # ..................{ INITIALIZORS                       }..................
    @type_check
    def __init__(self, phase: SimPhase) -> None:
        '''
        Initialize this simulation phase subcache.

        Parameters
        ----------
        phase : SimPhase
            Parent simulation phase.
        '''

        # Classify all passed parameters as weak rather than strong reference,
        # circumventing circular references and complications thereof.
        self._phase = pyref.proxy_weak(phase)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
CLI-specific functional tests exercising the equivalent circuit-based
simulation solver.
'''

# ....................{ IMPORTS                           }....................
# from betse.util.test.pytest.mark.pytfail import xfail

# ....................{ TESTS                             }....................
def test_cli_sim_fast(betse_cli_sim: 'CLISimTester') -> None:
    '''
    Functional test exporting all available exports (e.g., CSVs, plots,
    animations) with all simulation features required by these exports,
    including the equivalent circuit-based solver but excluding extracellular
    spaces (which are only partially supported by this solver).

    Parameters
    ----------
    betse_cli_sim : CLISimTester
        Object running BETSE CLI simulation subcommands.
    '''

    # Enable all exports and features required by these exports.
    betse_cli_sim.sim_state.config.enable_solver_fast_exports()

    # Test all default simulation-specific subcommands with this configuration.
    betse_cli_sim.run_subcommands_try()



#!/usr/bin/env python3
# ....................{ LICENSE                           }....................
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

# ....................{ IMPORTS                           }....................
import numpy as np
from betse.science import sim_toolbox as stb
from betse.science.math import finitediff as fd
from scipy.ndimage import gaussian_filter

# ....................{ GETTERS                           }....................
def get_current(sim, cells, p) -> None:
    '''
    Calculate currents in both cells and the environment.
    '''

    # calculate the net charge in cells:
    sim.rho_cells = np.dot(sim.zs * p.F, sim.cc_cells) + sim.extra_rho_cells

    # calculate membrane current density (- as fluxes were defined into cell)
    sim.Jmem = -np.dot(sim.zs * p.F, sim.fluxes_mem) + sim.extra_J_mem

    # calculate current density across cell membranes via gap junctions:
    sim.Jgj = np.dot(sim.zs * p.F, sim.fluxes_gj)

    # add the free current sources together into a single transmembrane current:
    sim.Jn = sim.Jmem + sim.Jgj

    # average the transmembrane current to the cell centre (for smoothing):
    Jn_ave = np.dot(cells.M_sum_mems, sim.Jn*cells.mem_sa) / cells.cell_sa
    # Smooth the free current at the membrane:
    sim.Jn = sim.smooth_weight_mem * sim.Jn + Jn_ave[cells.mem_to_cells] * sim.smooth_weight_o

    # multiply final result by membrane surface area to obtain current (direction into cell is +)
    sim.I_mem = -sim.Jn*cells.mem_sa

    # components of intracellular current:
    Jcx = sim.Jn * cells.mem_vects_flat[:,2]
    Jcy = sim.Jn * cells.mem_vects_flat[:,3]

    # average intracellular current to cell centres
    sim.J_cell_x = np.dot(cells.M_sum_mems, Jcx*cells.mem_sa) / cells.cell_sa
    sim.J_cell_y = np.dot(cells.M_sum_mems, Jcy*cells.mem_sa) / cells.cell_sa

    # normal component of J_cell at the membranes:
    sim.Jc = sim.J_cell_x[cells.mem_to_cells]*cells.mem_vects_flat[:,2] + sim.J_cell_y[cells.mem_to_cells]*cells.mem_vects_flat[:,3]

    # Current in the environment --------------------------------------------------------------------------------------
    if p.is_ecm is True:

        # calculate current densities in the environment from ion fluxes:
        J_env_x_o = np.dot(p.F*sim.zs, sim.fluxes_env_x) + sim.extra_Jenv_x
        J_env_y_o = np.dot(p.F*sim.zs, sim.fluxes_env_y) + sim.extra_Jenv_y

        # reshape the matrix:
        J_env_x_o = J_env_x_o.reshape(cells.X.shape)
        J_env_y_o = J_env_y_o.reshape(cells.X.shape)

        #---------------------------------------------
        # sigma_env = sim.sigma*sim.D_env_weight

        #Helmholtz-Hodge decomposition to obtain divergence-free projection of actual currents (zero n_hat at boundary):
        BB, sim.J_env_x, sim.J_env_y, _, Ja, Jb = stb.HH_Decomp(J_env_x_o, J_env_y_o, cells, bounds = sim.bound_V)

        # since it was incidentally calculated, store the B-component of magnetic field:
        sim.B_field = BB*p.mu

        # total free current density at each membrane:
        sim.Jtx = sim.J_env_x + Ja
        sim.Jty = sim.J_env_y + Jb


        # calculate the net charge in the environment:
        sim.rho_env = np.dot(sim.zs * p.F, sim.cc_env) + sim.extra_rho_env

        # Method 7-------------------------------------------------------------------------------------------------
        # Calculate voltage from charge in environment; this electric field is divergence-free assuming Laplace Eqn holds
        # This is the voltage measured with electrodes:

        # Boundary value problem:
        div_Jb = np.zeros(cells.X.shape)

        # add in extra boundary conditions for the case of an externally-applied voltage event:
        div_Jb[:, -1] = -sim.bound_V['R'] / cells.delta ** 2
        div_Jb[:, 0] = -sim.bound_V['L'] / cells.delta ** 2
        div_Jb[-1, :] = -sim.bound_V['T'] / cells.delta ** 2
        div_Jb[0, :] = -sim.bound_V['B'] / cells.delta ** 2

        Phi_b = np.dot(cells.lapENVinv, -div_Jb.ravel())

        # Voltage in the environment is related to extra surface surface charge:
        sim.rho_env_surf = np.zeros(sim.edl)
        sim.rho_env_surf[cells.map_mem2ecm] = sim.rho_env[cells.map_mem2ecm]*(p.cell_height*cells.delta**2)/(
            cells.memSa_per_envSquare[cells.map_mem2ecm].mean())

        sim.v_env = (sim.rho_env_surf / (sim.ko_env*p.eo*p.er))
        # sim.v_env = (sim.rho_env_surf / (p.cm))
        # sim.v_env = (sim.rho_env / (sim.ko_env**2 * p.eo * p.er))

        screen = (2/(sim.ko_env*cells.delta))*(p.cell_radius/p.true_cell_size)
        # screen = (2/(sim.ko_env*cells.delta))

        sim.v_env = gaussian_filter(sim.v_env.reshape(cells.X.shape), 1, mode='constant', cval=0.0).ravel() + Phi_b

        gVex, gVey = fd.gradient(screen*sim.v_env.reshape(cells.X.shape), cells.delta)

        sim.E_env_x = -gVex
        sim.E_env_y = -gVey

        sim.Eme = (sim.E_env_x.ravel()[cells.map_mem2ecm] * cells.mem_vects_flat[:, 2] +
               sim.E_env_y.ravel()[cells.map_mem2ecm] * cells.mem_vects_flat[:, 3])

        sim.Phi_b = Phi_b # Save the boundary voltage

    else:
        # divergence of current across membranes:
        vc = sim.vm/2

        vce = np.zeros(len(cells.xypts))
        # vce[cells.map_mem2ecm] = vc[cells.mem_to_cells] + divJm[cells.mem_to_cells]
        vce[cells.map_mem2ecm] = vc[cells.mem_to_cells]

        # Local field potential:
        Phi = -vce.reshape(cells.X.shape)

        # Smooth it:
        # Phi = fd.integrator(Phi.reshape(cells.X.shape), 0.5) # FIXME: Make this optional
        Phi = gaussian_filter(Phi, 1)

        # Calculate the gradient in free space (which is why this is multiplied by 'cells.delta):
        gVex, gVey = fd.gradient(cells.delta*Phi.reshape(cells.X.shape), cells.delta)

        # Smooth it:
        gVex, gVey = stb.smooth_flux(gVex.reshape(cells.X.shape), gVey.reshape(cells.X.shape), cells) # FIXME: Make this optional

        # Assign electric field:
        sim.E_env_x = -gVex
        sim.E_env_y = -gVey

        # Environmental current:
        Jxo = sim.sigma * sim.E_env_x * sim.D_env_weight
        Jyo = sim.sigma * sim.E_env_y * sim.D_env_weight

        BB, sim.J_env_x, sim.J_env_y, _, Ja, Jb = stb.HH_Decomp(Jxo, Jyo, cells)

        # since it was incidentally calculated, store the B-component of magnetic field:
        sim.B_field = BB * p.mu

        # total free current density at each membrane:
        sim.Jtx = sim.J_env_x + Ja
        sim.Jty = sim.J_env_y + Jb

        # map current from extracellular space to membrane normal
        sim.Eme = (sim.E_env_x.ravel()[cells.map_mem2ecm] * cells.mem_vects_flat[:, 2] +
               sim.E_env_y.ravel()[cells.map_mem2ecm] * cells.mem_vects_flat[:, 3])

        # assign environmental voltage:
        sim.v_env =  (Phi).ravel()

        # Boundary value problem:
        div_Jb = np.zeros(cells.X.shape)

        # add in extra boundary conditions for the case of an externally-applied voltage event:
        div_Jb[:, -1] = -sim.bound_V['R'] / cells.delta ** 2
        div_Jb[:, 0] = -sim.bound_V['L'] / cells.delta ** 2
        div_Jb[-1, :] = -sim.bound_V['T'] / cells.delta ** 2
        div_Jb[0, :] = -sim.bound_V['B'] / cells.delta ** 2

        Phi_b = np.dot(cells.lapENVinv, -div_Jb.ravel())
        sim.Phi_b = Phi_b # save the boundary value problem

# WASTELANDS (Options)--------------------------------------------------------------------------------------------------

    # sim.Phi_cells = ((sim.rho_cells) / ((sim.ko_cell ** 2) * p.eo * p.er))

    # gPhi = (sim.Phi_cells[cells.cell_nn_i[:, 1]] - sim.Phi_cells[cells.cell_nn_i[:, 0]]) / (cells.nn_len)

    # Electric field in the cells:
    # Exm = -gPhi*cells.mem_vects_flat[:,2]
    # Eym = -gPhi*cells.mem_vects_flat[:,3]

    # average intracellular electric field at cell centres:
    # sim.E_cell_x = np.dot(cells.M_sum_mems, Exm*cells.mem_sa) / cells.cell_sa
    # sim.E_cell_y = np.dot(cells.M_sum_mems, Eym*cells.mem_sa) / cells.cell_sa

        # Method 2: Integrated charge calculation from currents:-------------------------------------------------------

        #divergence of the environmental current from fluxes in the environment:
        # div_Je = fd.divergence(sim.Jtx, sim.Jty, cells.delta, cells.delta)
        #
        # #divergence of trans-membrane fluxes:
        # div_Je_fromcells = stb.div_env(-sim.Jmem, cells, p).reshape(cells.X.shape)
        # div_Je += div_Je_fromcells
        #
        # #calculate mapped current component from transmembrane fluxes (which is always curl-free):
        # #Important for 100% biophysical correctness, but we can skip adding these in for efficiency
        # Phi_cells = np.dot(cells.lapENVinv, -div_Je_fromcells.ravel())
        # Jex_cells, Jey_cells = fd.gradient(Phi_cells.reshape(cells.X.shape), cells.delta)
        #
        # sim.Jtx += Jex_cells
        # sim.Jty += Jey_cells
        #
        # # #the negative divergence of the total environmental current is the change in charge density:
        # sim.rho_env += -div_Je.ravel()*p.dt


    # ---Method # 1----------------------------------------------------------------------------------------
    # Handle the electric field in the environment using the Screened Poisson equation for currents:

    # divJ = fd.divergence(sim.Jtx, sim.Jty, cells.delta, cells.delta)
    #
    # # lambda_screen = 1e-9
    # v_env = sim.v_env.reshape(cells.X.shape)
    # v_env += -(divJ /((sim.ko_env**2)*p.er*p.eo))*p.dt
    #
    # # if p.sharpness < 1.0:
    # v_env = fd.integrator(v_env, 0.5)
    #
    # # add in extra boundary conditions for the case of an externally-applied voltage event:
    # v_env[:, -1] = sim.bound_V['R']
    # v_env[:, 0] = sim.bound_V['L']
    # v_env[-1, :] = sim.bound_V['T']
    # v_env[0, :] = sim.bound_V['B']
    #
    # gVex, gVey = fd.gradient(v_env, cells.delta)
    #
    # # assign to environmental voltage array:
    # sim.v_env = 1*v_env.ravel()
    #
    # # use Hodgkin-Huxley decomposition and re-composition to "smooth" the electric field:
    # gVex, gVey = stb.smooth_flux(gVex.reshape(cells.X.shape), gVey.reshape(cells.X.shape), cells)
    #
    # # assign to electric field of the system:
    # sim.E_env_x = -gVex
    # sim.E_env_y = -gVey
    #
    # sim.Eme = (sim.E_env_x.ravel()[cells.map_mem2ecm] * cells.mem_vects_flat[:, 2] +
    #            sim.E_env_y.ravel()[cells.map_mem2ecm] * cells.mem_vects_flat[:, 3])

    # --Method #2 ---------------------------------------------------------------

    # The solution to the Screened Poisson Equation in the limit of large screening constant Ko, is simply
    # Phi = +f/Ko2. This makes a perfect voltage estimate for the extracellular space.
    # (the Screened Poisson Equation is Lap(Phi) - ko2 Phi = -rho/(eta)
    # Note that the relative permittivity of the double layer is known to be 6 rather than 80 of pure water
    # (see Srinivasan 2006).

    # v_env = ((sim.rho_env) / ((sim.ko_env ** 2) * p.eo * p.er))
    # v_env = v_env.reshape(cells.X.shape)
    #
    # v_env = fd.integrator(v_env, 0.5)
    #
    # # add in extra boundary conditions for the case of an externally-applied voltage event:
    # v_env[:, -1] = sim.bound_V['R']
    # v_env[:, 0] = sim.bound_V['L']
    # v_env[-1, :] = sim.bound_V['T']
    # v_env[0, :] = sim.bound_V['B']
    #
    # # gradient of the polarization voltage yields the electric field:
    # gVex, gVey = fd.gradient(v_env, cells.delta)
    #
    # # assign to environmental voltage array:
    # sim.v_env = 1*v_env.ravel()
    #
    # # use Hodgkin-Huxley decomposition and recomposition to "smooth" the electric field:
    # gVex, gVey = stb.smooth_flux(gVex.reshape(cells.X.shape), gVey.reshape(cells.X.shape), cells)
    #
    # # assign to electric field of the system:
    # sim.E_env_x = -gVex
    # sim.E_env_y = -gVey
    #
    #
    # sim.Eme = (sim.E_env_x.ravel()[cells.map_mem2ecm] * cells.mem_vects_flat[:, 2] +
    #        sim.E_env_y.ravel()[cells.map_mem2ecm] * cells.mem_vects_flat[:, 3])

    # --Method 3: Local field potentials--------------------------------------------------------------------------

    # env_sig = sim.sigma * sim.D_env_weight # environmental conductivity map
    #
    # # env_sig = sim.sigma_env.reshape(cells.X.shape) # environmental conductivity map
    # # divergence of environmental current, scaled by the conductivity map
    # divJo = fd.divergence(sim.Jtx/env_sig, sim.Jty/env_sig, cells.delta, cells.delta)
    #
    # # set boundary conditions for any applied voltages:
    # divJo[:, -1] = -sim.bound_V['R'] / (cells.delta ** 2)
    # divJo[:, 0] = -sim.bound_V['L'] / (cells.delta ** 2)
    # divJo[-1, :] = -sim.bound_V['T'] / (cells.delta ** 2)
    # divJo[0, :] = -sim.bound_V['B'] / (cells.delta ** 2)
    #
    # # divJo = fd.integrator(divJo.reshape(cells.X.shape), 0.5)
    #
    # # environmental local field potential:
    # Phi = np.dot(cells.lapENVinv, -divJo.ravel())
    #
    # # smooth it:
    # Phi = fd.integrator(Phi.reshape(cells.X.shape), 0.5)
    #
    # # take the gradient:
    # gVex, gVey = fd.gradient(Phi.reshape(cells.X.shape), cells.delta)
    #
    # # use Hodgkin-Huxley decomposition and recomposition to "smooth" the electric field:
    # gVex, gVey = stb.smooth_flux(gVex.reshape(cells.X.shape), gVey.reshape(cells.X.shape), cells)
    #
    # # assign to electric field of the system:
    # sim.E_env_x = -gVex
    # sim.E_env_y = -gVey
    #
    # sim.Eme = (sim.E_env_x.ravel()[cells.map_mem2ecm] * cells.mem_vects_flat[:, 2] +
    #        sim.E_env_y.ravel()[cells.map_mem2ecm] * cells.mem_vects_flat[:, 3])
    #
    # # assign environmental voltage:
    # sim.v_env = Phi.ravel() * 1

    # # Method 4------------------------------------------------------------------------------------------------
    # # Charge in the environmental space is surface charge that screens the cell charge.
    #
    # # Voltage seen in the extracellular region:
    # vce = np.zeros(sim.edl)
    # vce[cells.map_mem2ecm] = -sim.vm/2
    #
    # # env_sig = sim.sigma * sim.D_env_weight # environmental conductivity map
    # # # divergence of environmental current, scaled by the conductivity map
    # # divJo = fd.divergence(sim.Jtx/env_sig, sim.Jty/env_sig, cells.delta, cells.delta)
    # #
    # # # set boundary conditions for any applied voltages:
    # # divJo[:, -1] = -sim.bound_V['R'] / (cells.delta ** 2)
    # # divJo[:, 0] = -sim.bound_V['L'] / (cells.delta ** 2)
    # # divJo[-1, :] = -sim.bound_V['T'] / (cells.delta ** 2)
    # # divJo[0, :] = -sim.bound_V['B'] / (cells.delta ** 2)
    # #
    # # # environmental local field potential:
    # # lfp = np.dot(cells.lapENVinv, -divJo.ravel())
    #
    # # v_env = vce + ((sim.rho_env) / ((sim.ko_env ** 2) * p.eo * p.er))
    #
    # v_env = vce
    #
    # v_env = v_env.reshape(cells.X.shape)
    #
    # v_env = fd.integrator(v_env, 0.5)
    #
    # # assign to environmental voltage array:
    # # sim.v_env = 1*v_env.ravel()
    #
    # # gradient of the polarization voltage yields the electric field:
    # gVex, gVey = fd.gradient(v_env.reshape(cells.X.shape), cells.delta)
    #
    # # use Hodgkin-Huxley decomposition and recomposition to "smooth" the electric field:
    # # gVex, gVey = stb.smooth_flux(gVex.reshape(cells.X.shape), gVey.reshape(cells.X.shape), cells)
    # sim.v_env, gVex, gVey, _, _, _ = stb.HH_Decomp(gVex, gVey, cells)
    #
    # # assign to electric field of the system:
    # sim.E_env_x = -gVex
    # sim.E_env_y = -gVey
    #
    # sim.Eme = (sim.E_env_x.ravel()[cells.map_mem2ecm] * cells.mem_vects_flat[:, 2] +
    #        sim.E_env_y.ravel()[cells.map_mem2ecm] * cells.mem_vects_flat[:, 3])

    # Method 6-------------------------------------------------------------------------------------------------
    # Calculate voltage from charge in environment; this electric field is divergence-free assuming Laplace Eqn holds
    # This is the voltage measured with electrodes:
    # v_env = np.zeros(sim.edl)
    #
    # v_env[cells.map_mem2ecm] = (1 / (2 * p.cm)) * (sim.rho_env[cells.map_mem2ecm] * cells.delta ** 2 * p.cell_height
    #                                                ) / (cells.memSa_per_envSquare[cells.map_mem2ecm].mean())
    #
    # v_env = v_env.reshape(cells.X.shape)
    #
    # v_env = fd.integrator(v_env, 0.5)  # FIXME! Make this optional
    #
    # # assign to environmental voltage array:
    # sim.v_env = 1 * v_env.ravel()
    #
    # # Now calculate the voltage that results from time-dependent charge transfers
    # # these induce movements of charges and other factors.
    # Phi_env = ((sim.rho_env) / ((sim.ko_env ** 2) * p.eo * p.er))
    #
    # Phi_env = fd.integrator(Phi_env.reshape(cells.X.shape), 0.5)  # FIXME! Make this optional
    #
    # # gradient of the polarization voltage yields the electric field:
    # gVex, gVey = fd.gradient(Phi_env.reshape(cells.X.shape), cells.delta)
    #
    # # use Hodgkin-Huxley decomposition and recomposition to "smooth" the electric field:
    # gVex, gVey = stb.smooth_flux(gVex.reshape(cells.X.shape), gVey.reshape(cells.X.shape),
    #                              cells)  # FIXME! Make this optional
    #
    # # assign to electric field of the system:
    # sim.E_env_x = -gVex
    # sim.E_env_y = -gVey
    #
    # sim.Eme = (sim.E_env_x.ravel()[cells.map_mem2ecm] * cells.mem_vects_flat[:, 2] +
    #            sim.E_env_y.ravel()[cells.map_mem2ecm] * cells.mem_vects_flat[:, 3])
    #
    # sim.Phi_env = Phi_env * 1



#!/usr/bin/env python3
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Wound-induced transient channel response classes.
'''

# .................... IMPORTS                            ....................
from abc import ABCMeta, abstractmethod
import numpy as np
from betse.exceptions import BetseSimConfException
from betse.science import sim_toolbox as stb
from betse.science.channels.channelsabc import ChannelsABC
from betse.science.math import toolbox as tb
from betse.util.io.log import logs
# from betse.science.chemistry.molecule import get_influencers

# .................... BASE                               ....................
class WoundABC(ChannelsABC, metaclass=ABCMeta):
    '''
    Abstract base class of all wound-induced (TRP-based) channel classes.

    Attributes
    ----------
    _mInf :
        Equillibrium value function for m-gates of channel
    _mTau :
        Time-constant function for m-gates of channel
    _hInf :
        Equillibrium value function for h-gates of channel
    _hTau :
        Time-constant function for h-gates of channel

    '''

    def init(self, vm, cells, p, targets = None):
        '''
        Initialize wound-induced (TRP based) channel at the point of wounding.

        '''

        if targets is None:

            self.targets = cells.mem_i

        else:
            self.targets = targets

        self.data_length = len(self.targets)
        self.mdl = len(cells.mem_i)

        self.modulator = 1.0

        self.v_corr = 0.0   # in experiments, the measurement junction voltage is about 10 mV

        V = vm[self.targets]*1000 + self.v_corr
        # V = vm * 1000 + self.v_corr

        self._init_state(V)

        self.ions = ['Na', 'K']
        self.rel_perm = [1.0, 1.0]

        self.W_factor = 2.0   # initial concentration of the 'wound factor' (which is basically pressure)
        self.W_decay = p.wound_close_factor


    def run(self, vm, p):
        '''
        Simulate wound-induced (TRP based) channel activity after wounding.

        '''

        V = vm[self.targets]*1000

        self._calculate_state(V)

        self.W_factor = self.W_factor - self.W_factor*self.W_decay*p.dt

        self._implement_state(V, p)

    def _implement_state(self, V, p):
        # Update the 'm' and 'h' gating functions:
        self.update_mh(p, time_unit = self.time_unit)

        # calculate the open-probability of the channel:
        P = (self.m ** self._mpower) * (self.h ** self._hpower)

        self.P = np.zeros(self.mdl)
        self.P[self.targets] = P


        # # obtain concentration of ion inside and out of the cell, as well as its charge z:
        # c_mem_Na = sim.cc_cells[sim.iNa][cells.mem_to_cells]
        # c_mem_K = sim.cc_cells[sim.iK][cells.mem_to_cells]
        #
        # if p.is_ecm is True:
        #     c_env_Na = sim.cc_env[sim.iNa][cells.map_mem2ecm]
        #     c_env_K = sim.cc_env[sim.iK][cells.map_mem2ecm]
        #
        # else:
        #     c_env_Na = sim.cc_env[sim.iNa]
        #     c_env_K = sim.cc_env[sim.iK]
        #
        # IdM = np.ones(sim.mdl)
        #
        # z_Na = sim.zs[sim.iNa] * IdM
        # z_K = sim.zs[sim.iK] * IdM
        #
        # # membrane diffusion constant of the channel:
        # Dchan = dyna.maxDmWound*P*(self.W_factor/(1 + self.W_factor))*self.modulator
        #
        # # calculate specific ion flux contribution for this channel:
        # delta_Q_Na = stb.electroflux(c_env_Na, c_mem_Na, Dchan, p.tm * IdM, z_Na, sim.vm, sim.T, p, rho=sim.rho_channel)
        # delta_Q_K = stb.electroflux(c_env_K, c_mem_K, Dchan, p.tm * IdM, z_K, sim.vm, sim.T, p, rho=sim.rho_channel)
        #
        # self.clip_flux(delta_Q_Na, threshold=p.flux_threshold)
        # self.clip_flux(delta_Q_K, threshold=p.flux_threshold)
        #
        # self.chan_flux = np.zeros(sim.mdl)
        # self.chan_flux[dyna.targets_vgWound] = -delta_Q_Na[dyna.targets_vgWound] -delta_Q_K[dyna.targets_vgWound]
        #
        # self.update_charge(sim.iNa, delta_Q_Na, dyna.targets_vgWound, sim, cells, p)
        # self.update_charge(sim.iK, delta_Q_K, dyna.targets_vgWound, sim, cells, p)

        # if p.ions_dict['Ca'] == 1.0:
        #     self.update_charge(sim.iCa, 0.1*delta_Q, dyna.targets_vgWound, sim, cells, p)


    @abstractmethod
    def _init_state(self, V, sim, p):
        '''
        Initializes values of the m and h gates of the channel.
        '''
        pass


    @abstractmethod
    def _calculate_state(self, V, sim, p):
        '''
        Calculates time-dependent values of the m and h gates of the channel.
        '''
        pass

# ....................{ SUBCLASS                           }....................
class TRP(WoundABC):
    '''
    TRP model default.

    '''

    def _init_state(self, V, sim, p):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using the wound-induced channel: TRP')

        self.v_corr = 0

        # initialize values of the m and h gates of the channel on m_inf and h_inf:
        self.m = np.ones(sim.mdl)
        self.h = np.ones(sim.mdl)

        # define the power of m and h gates used in the final channel state equation:
        self._mpower = 0
        self._hpower = 0


    def _calculate_state(self, V, sim, p):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """

        self.vrev = 0  # reversal voltage used in model [mV]

        self._mInf = 1.0
        self._mTau = 1.0
        self._hInf = 1.0
        self._hTau = 1.0





#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Fixture classes encapsulating test-related simulation configurations.
'''

# ....................{ TODO                               }....................
#FIXME: Most use of the increasingly obsolete "SimConfTestInternal.config"
#wrapper attribute (both here and everywhere else) should be replaced by use of
#the new "SimConfTestInternal.p" property, which increasingly provides all test
#functionality.

# ....................{ IMPORTS                            }....................
from abc import ABCMeta
from betse.util.type.decorator.deccls import abstractproperty
from betse.util.type.types import type_check

# ....................{ SUPERCLASSES                       }....................
class SimConfTestABC(object, metaclass=ABCMeta):
    '''
    Abstract base class of all simulation configuration context subclasses,
    each encapsulating state and metadata for a test-related simulation
    configuration.

    Simulation configuration fixtures typically return instances of this class
    as a means of communicating this context to other fixtures and tests.

    Attributes
    ----------
    conf_dirname : str
        Absolute dirname of the directory of the simulation configuration file
        wrapped by this wrapper (i.e., of :attr:`conf_filename`).
    conf_filename : str
        Absolute filename of the simulation configuration file wrapped by this
        wrapper.

    See Also
    ----------
    https://py.readthedocs.org/en/latest/path.html
        Official :mod:`py.path` submodule documentation.
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def __init__(self, conf_filename: str) -> None:
        '''
        Initialize this simulation configuration context.

        Parameters
        ----------
        conf_filename : str
            Absolute filename of the simulation configuration file encapsulated
            by this context. Although this file need *not* physically exist
            before this object is instantiated (i.e., before this method is
            called), the subclass should ensure this file exists once this
            object is instantiated (i.e., immediately before the subclass
            implementation of this method returns).
        '''

        # Defer heavyweight imports.
        from betse.util.path import pathnames

        # Classify all passed parameters publicly, permitting access by tests.
        self.conf_filename = conf_filename
        self.conf_dirname = pathnames.get_dirname(self.conf_filename)

    # ..................{ CONTEXTS                           }..................
    def context(self) -> 'contextlib.contextmanager':
        '''
        Context manager changing the current working directory (CWD) of the
        current test to the directory containing this configuration file for
        the duration of this context.

        Default simulation configuration paths are relative to the directory
        containing the simulation configuration file: namely, this temporary
        directory. Changing directories resolves these paths to this directory.
        Failing to do so would incorrectly resolve these paths to the current
        directory, with predictably disastrous outcomes. While this class could
        instead globally search-and-replace all relative simulation
        configuration paths with absolute paths, doing so would be considerably
        more complex, fragile, and error-prone than changing directories.
        '''

        # Defer heavyweight imports.
        from betse.util.os.shell import shelldir

        # Defer to the generator returned by the following utility function.
        return shelldir.setting_cwd(self.conf_dirname)

    # ..................{ SUBCLASS                           }..................
    # Subclasses are required to define the following read-only properties.

    @abstractproperty
    def p(self) -> 'betse.science.parameters.Parameters':
        '''
        High-level simulation configuration encapsulated by this test wrapper.
        '''

        pass

# ....................{ SUBCLASSES                         }....................
class SimConfTestInternal(SimConfTestABC):
    '''
    Simulation configuration context subclass encapsulating a temporary
    simulation configuration file internally isolated to the current test.

    Since this configuration is guaranteed to be test-specific and hence safely
    modifiable, caller fixtures and tests may safely modify the contents of this
    configuration (e.g., to exercise specific feature sets and edge cases).

    Attributes
    ----------
    config : SimConfigWrapper
        Simulation configuration wrapper wrapping the low-level dictionary
        deserialized from the YAML-formatted simulation configuration file with
        path :attr:`conf_filepath`. Note the contents of this dictionary may be
        desynchronized from those of this file. For efficiency, callers may
        modify this dictionary to suite test requirements *before*
        reserializing this dictionary back to this file.
    conf_filepath : py._path.local.LocalPath
        Absolute filename of a temporary simulation configuration file specific
        to the parent fixture as a :class:`py.path.local` instance, defining an
        object-oriented superset of the non-object-oriented :mod:`os.path`
        module.

    See Also
    ----------
    https://py.readthedocs.org/en/latest/path.html
        Official :mod:`py.path` submodule documentation.
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def __init__(
        self,
        src_conf_filename: str,
        trg_conf_filepath: 'py._path.local.LocalPath',
    ) -> None:
        '''
        Initialize this context by copying the passed source to target
        simulation configuration file.

        Specifically, this method (in order):

        #. Copies the passed source to target simulation configuration file.
        #. Copies *all* external resources (e.g., image files) referenced and
           hence required by the passed source simulation configuration file to
           the directory of the passed target simulation configuration file.
        #. Sanitizes the copied simulation configuration file for all child
           fixtures and tests by unconditionally disabling options either
           requiring interactive input *or* displaying interactive output.

        This method does *not* minimize the space and time costs of running
        simulations configured by this configuration. Doing so is typically
        desirable but can obscure edge-case issues, including computational
        instability produced by the default non-minified time steps.

        Caveats
        ----------
        For efficiency, note that the configuration changes applied by this
        method (listed above) reside *only* in-memory; they have yet to be
        written back to disk. Callers are required to do so manually (e.g., via
        the optional ``is_overwrite_conf`` parameter passed to the
        :meth:`CLISimTester.run_subcommands` method).

        Parameters
        ----------
        src_conf_filename : str
            Absolute filename of the source simulation configuration file from
            which this method safely copies the passed target simulation
            configuration file to.
        trg_conf_filepath : py._path.local.LocalPath
            Absolute filename of the target simulation configuration file to
            which this method safely copies both the passed source simulation
            configuration file *and* all external resources (e.g., image files)
            referenced and hence required by that file. For compliance with
            :mod:`pytest` internals, this filename is a high-level
            :class:`py.path.local` instance rather than a low-level string.

        Raises
        ----------
        BetseDirException
            If any external file required by this target file already exists.
        BetseFileException
            If this target file already exists.
        '''

        # Defer heavyweight imports. Notably, the "simconfwrapper" submodule
        # inherits from the main codebase and is thus *NOT* safely importable
        # at module scope.
        from betse.science.parameters import Parameters
        from betse_test._fixture.simconf.simconfwrapper import (
            SimConfigTestWrapper)

        # Absolute filename of this file.
        trg_conf_filename = str(trg_conf_filepath)

        # Initialize our superclass with this filename.
        super().__init__(conf_filename=trg_conf_filename)

        # Classify the passed parameters. While the "self.config" object
        # classified below provides this filename as a low-level string, this
        # high-level "py.path.local" instance is useful in fixtures and tests.
        self.conf_filepath = trg_conf_filepath

        # Copy this source to target file.
        p = Parameters()
        p.copy(
            src_conf_filename=src_conf_filename,
            trg_conf_filename=trg_conf_filename)

        # Test-specific wrapper encapsulating this file.
        self.config = SimConfigTestWrapper(p)

        # Sanitize this configuration for all child fixtures and tests.
        self.config.disable_interaction()

    # ..................{ SUPERCLASS                         }..................
    @property
    def p(self) -> 'betse.science.parameters.Parameters':
        return self.config.p


class SimConfTestExternal(SimConfTestABC):
    '''
    Simulation configuration context subclass encapsulating an external
    simulation configuration file residing at any directory and hence *not*
    necessarily isolated to the current test.

    Attributes
    ----------
    _p : Parameters
        High-level simulation configuration encapsulating this external
        simulation configuration file.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(self, *args, **kwargs) -> None:
        '''
        Initialize this simulation configuration context.
        '''

        # Defer heavyweight imports.
        from betse.science.parameters import Parameters

        # Initialize our superclass with all passed parameters.
        super().__init__(*args, **kwargs)

        # In-memory simulation configuration deserialized from this file.
        self._p = Parameters.make(conf_filename=self.conf_filename)

    # ..................{ SUPERCLASS                         }..................
    @property
    def p(self) -> 'betse.science.parameters.Parameters':
        return self._p



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Application-specific exception hierarchy.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To avoid race conditions during application startup, this module may
# import *ONLY* from modules guaranteed to exist at startup. This includes all
# standard Python and application modules but *NOT* third-party dependencies,
# which if unimportable will only be validated at some later time in startup.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from abc import ABCMeta

# ....................{ EXCEPTIONS                         }....................
#FIXME: Define an __init__() method asserting that the passed exception message
#is non-None, which Python permits by default but which is functionally useless.
class BetseException(Exception, metaclass=ABCMeta):
    '''
    Abstract base class of all application-specific exceptions.
    '''

    pass


class BetseAttrException(BetseException):
    '''
    **Attribute** (i.e., variable or method bound to an object)-specific
    exception.
    '''

    pass


class BetseMetaAppException(BetseException):
    '''
    **Application metadata singleton** (i.e., application-wide object
    synopsizing application metadata via read-only properties)-specific
    exception.
    '''

    pass

# ....................{ EXCEPTIONS ~ call                  }....................
class BetseCallableException(BetseException):
    '''
    General-purpose exception applicable to all **callables** (e.g., functions,
    lambdas, methods, properties).
    '''

    pass


class BetseCallbackException(BetseCallableException):
    '''
    Callback-specific exception.
    '''

    pass


class BetseDecoratorException(BetseCallableException):
    '''
    Decorator-specific exception.
    '''

    pass


class BetseLambdaException(BetseCallableException):
    '''
    Lambda-specific exception.
    '''

    pass


class BetseParamException(BetseCallableException):
    '''
    **Parameter** (i.e., positional or keyword argument passed to a
    callable)-specific exception.
    '''

    pass

# ....................{ EXCEPTIONS ~ call : func           }....................
class BetseFunctionException(BetseCallableException):
    '''
    Function-specific exception.
    '''

    pass


class BetseFunctionUnimplementedException(
    BetseFunctionException, NotImplementedError):
    '''
    Unimplemented function-specific exception.

    This exception is typically raised from **unimplemented dependency- or
    platform-specific functions** (i.e., functions assuming a specific
    dependency or platform that have yet to be implemented).
    '''

# ....................{ EXCEPTIONS ~ call : method         }....................
class BetseMethodException(BetseCallableException):
    '''
    Method-specific exception.
    '''

    pass


class BetseMethodUnimplementedException(
    BetseMethodException, NotImplementedError):
    '''
    Unimplemented method-specific exception.

    This exception is typically raised from **unimplemented optional methods**
    (i.e., non-mandatory methods *not* intended to be called) of concrete
    subclasses of abstract base classes. While the optimal solution for
    defining **unimplemented mandatory methods** (i.e., non-optional methods
    also *not* intended to be called) is via the standard
    :call:`collections.abc.abstractmethod` decorator, there exists no standard
    alternative for defining optional methods. Hence, this exception.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(self) -> None:

        # Avoid circular import dependencies.
        from betse.util.type.call import callers

        # Initialize our superclass with a human-readable exception message.
        super().__init__('Method {}() unimplemented.'.format(
            callers.get_caller_basename()))

# ....................{ EXCEPTIONS ~ call : descriptor     }....................
class BetseDescriptorException(BetseCallableException):
    '''
    Descriptor-specific exception.
    '''
    pass


class BetseExprAliasException(BetseDescriptorException):
    '''
    Expression alias-specific exception.
    '''
    pass

# ....................{ EXCEPTIONS ~ cli                   }....................
class BetseCLIException(BetseException):
    '''
    General-purpose command-line interface (CLI) exception.
    '''

    pass


class BetseCLIArgException(BetseCLIException):
    '''
    Command-line interface (CLI) argument-specific exception.
    '''

    pass


class BetseCLIArgParserException(BetseCLIArgException):
    '''
    **Argument parser** (i.e., :class:`argparse.ArgumentParser`)-specific
    exception.
    '''

    pass

# ....................{ EXCEPTIONS ~ lib                   }....................
class BetseLibException(BetseException):
    '''
    General-purpose exception applicable to third-party dependencies.
    '''

    pass


class BetseMatplotlibException(BetseLibException):
    '''
    :mod:`matplotlib`-specific exception.
    '''

    pass


class BetsePyDotException(BetseLibException):
    '''
    :mod:`pydot`-specific exception.
    '''

    pass


class BetseVersionException(BetseLibException):
    '''
    Version specifier-specific exception applicable to third-party dependency.
    versions.
    '''

    pass


class BetseYamlException(BetseException):
    '''
    Yet Another Markup Language (YAML)-specific exception.
    '''

    pass

# ....................{ EXCEPTIONS ~ log                   }....................
class BetseLogException(BetseException):
    '''
    Logging-specific exception.
    '''

    pass


class BetseLogRaceException(BetseLogException):
    '''
    Logging race condition-specific exception.

    Motivation
    ----------
    Concurrent attempts to log to the same physical file from multiple
    processes can and typically will produce fatal race conditions producing
    raised exceptions from one or more of these processes. On logfile rotation,
    each process will aggressively contend with each other process for write
    access to the same physical file to be rotated.

    The thread- *and* process-safe
    :class:`betse.util.io.log.conf.logconfhandle.RotatingFileHandler` handler
    obviates most of these concerns. Unfortunately, due to inadequacies in the
    Python ecosystem with respect to process-safe logging, that handler
    *cannot* constrain logfile access during rotation. Instead, on detecting
    exceptions produced by race conditions between multiple processes competing
    for access when attempting to emit log records, that handler temporarily
    halts the existing process for a negligible amount of the timeslice (e.g.,
    100ms) and repeats the attempt a negligible number of times (e.g., 8)
    *before* giving up and raising a fatal instance of this exception class.
    '''

    pass

# ....................{ EXCEPTIONS ~ math                  }....................
class BetseMathException(BetseException):
    '''
    General-purpose exception applicable to all low-level math algorithms.
    '''

    pass


class BetseMathInterpolationException(BetseMathException):
    '''
    Interpolation-specific math exception.
    '''

    pass


class BetseMathLineException(BetseMathException):
    '''
    Line- and line segment-specific math exception.
    '''

    pass


class BetseMathMeshException(BetseMathException):
    '''
    Mesh-specific math exception.
    '''

    pass


class BetseMathPointException(BetseMathException):
    '''
    Point-specific math exception.
    '''

    pass


class BetseMathPolygonException(BetseMathException):
    '''
    Polygon-specific math exception.
    '''

    pass

# ....................{ EXCEPTIONS ~ module                }....................
class BetseModuleException(BetseException):
    '''
    Module-specific exception.
    '''

    pass


class BetsePackageException(BetseModuleException):
    '''
    Module-specific exception.
    '''

    pass

# ....................{ EXCEPTIONS ~ os                    }....................
class BetseOSException(BetseException):
    '''
    General-purpose low-level operating system (OS) exception.
    '''

    pass


class BetseShellEnvException(BetseOSException):
    '''
    Shell environment-specific exception.
    '''

    pass

# ....................{ EXCEPTIONS ~ os : proc             }....................
class BetseProcessException(BetseOSException):
    '''
    General-purpose low-level process exception.
    '''

    pass


class BetseProcessNotFoundException(BetseProcessException):
    '''
    Low-level exception raised on failing to find a given process.
    '''

    pass

# ....................{ EXCEPTIONS ~ path                  }....................
class BetsePathnameException(BetseException):
    '''
    Pathname-specific exception.
    '''

    pass


class BetsePathException(BetseException):
    '''
    Path-specific exception.
    '''

    pass


class BetseArchiveException(BetsePathException):
    '''
    Archive-specific exception.
    '''

    pass


class BetseDirException(BetsePathException):
    '''
    Directory-specific exception.
    '''

    pass


class BetseGitException(BetsePathException):
    '''
    Git-specific exception.
    '''

    pass

# ....................{ EXCEPTIONS ~ path : file           }....................
class BetseFileException(BetsePathException):
    '''
    File-specific exception.
    '''

    pass


class BetseCommandException(BetseFileException):
    '''
    **Command** (i.e., executable binary file)-specific exception.
    '''

    pass


class BetseImageException(BetseFileException):
    '''
    Image file-specific exception.
    '''

    pass

# ....................{ EXCEPTIONS ~ py                    }....................
class BetsePyException(BetseException):
    '''
    General-purpose low-level Python interpreter exception.

    This exception is appropriate for use in relation to low-level issues
    concerning the active Python interpreter (e.g., inability to retrieve this
    interpreter's absolute path).
    '''

    pass


class BetsePyFrozenException(BetsePyException):
    '''
    **Frozen executable** (i.e., Python codebase converted into a
    platform-specific executable)-specific exception.
    '''

    pass


class BetsePyIdentifierException(BetsePyException):
    '''
    **Python identifier** (i.e., class, module, or attribute name)-specific
    exception.
    '''

    pass

# ....................{ EXCEPTIONS ~ test                  }....................
class BetseTestException(BetseException):
    '''
    General-purpose exception pertaining to this application's test suite.
    '''

    pass


class BetseTestFixtureException(BetseTestException):
    '''
    Fixture-specific test exception.
    '''

    pass


class BetseTestHookException(BetseTestException):
    '''
    Hook-specific test exception.
    '''

    pass


class BetseTestParamException(BetseTestException):
    '''
    Parameter-specific test exception.
    '''

    pass

# ....................{ EXCEPTIONS ~ type                  }....................
class BetseTypeException(BetseException):
    '''
    General-purpose exception applicable to all types (i.e., classes).
    '''

    pass


class BetseEnumException(BetseTypeException):
    '''
    Enumeration-specific exception.
    '''

    pass


class BetseNumericException(BetseTypeException):
    '''
    **Numeric** (i.e., integer *and* float)-specific exception.
    '''

    pass


class BetseIntException(BetseTypeException):
    '''
    Integer-specific exception.
    '''

    pass

# ....................{ EXCEPTIONS ~ type : iter           }....................
class BetseIterableException(BetseTypeException):
    '''
    Iterable-specific exception.
    '''

    pass


class BetseSequenceException(BetseIterableException):
    '''
    Sequence-specific exception.
    '''

    pass

# ....................{ EXCEPTIONS ~ type : iter : map     }....................
class BetseMappingException(BetseIterableException):
    '''
    Dictionary-specific exception.
    '''

    pass


class BetseMappingKeyException(BetseMappingException):
    '''
    Dictionary key-specific exception.
    '''

    pass


class BetseMappingValueException(BetseMappingException):
    '''
    Dictionary value-specific exception.
    '''

    pass

# ....................{ EXCEPTIONS ~ type : str            }....................
class BetseStrException(BetseTypeException):
    '''
    String-specific  exception.
    '''

    pass


class BetseCharException(BetseStrException):
    '''
    **Character** (i.e., string of length 1)-specific  exception.
    '''

    pass



class BetseRegexException(BetseStrException):
    '''
    Regular exception-specific  exception.
    '''

    pass

# ....................{ EXCEPTIONS ~ sci : sim             }....................
class BetseSimException(BetseException):
    '''
    General-purpose simulation exception.
    '''

    pass


class BetseSimConfException(BetseSimException):
    '''
    Simulation configuration-specific exception.
    '''

    pass


class BetseSimEventException(BetseSimException):
    '''
    Simulation event-specific exception.
    '''

    pass


class BetseSimPhaseException(BetseSimException):
    '''
    Simulation phase-specific exception.
    '''

    pass


class BetseSimTissueException(BetseSimException):
    '''
    Simulation tissue-specific exception.
    '''

    pass


class BetseSimVectorException(BetseSimException):
    '''
    Vector-specific exception.
    '''

    pass

# ....................{ EXCEPTIONS ~ sci : sim : unstable  }....................
class BetseSimUnstableException(BetseSimException):
    '''
    Simulation-specific exception indicating the current simulation to have
    unexpectedly failed due to computational instability.
    '''

    pass


class BetseSimUnstableNaNException(BetseSimUnstableException):
    '''
    Simulation-specific exception indicating the current simulation to have
    unexpectedly failed due to a **NaN-based computational instability** (i.e.,
    an operation producing a Numpy array containing at least one NaN value,
    typically due to division by zero or infinity).
    '''

    pass

# ....................{ EXCEPTIONS ~ sci : visual          }....................
class BetseSimVisualException(BetseSimException):
    '''
    Simulation visualization-specific exception, applicable to both plots and
    animations.
    '''

    pass


class BetseSimVisualLayerException(BetseSimVisualException):
    '''
    Simulation visualization layer-specific exception.
    '''

    pass

# ....................{ EXCEPTIONS ~ sci : pipe            }....................
class BetseSimPipeException(BetseSimException):
    '''
    Simulation pipeline-specific exception.
    '''

    pass


class BetseSimPipeRunnerUnsatisfiedException(BetseSimPipeException):
    '''
    Simulation pipeline-specific exception raised on attempting to run a runner
    with unsatisfied requirements (e.g., a post-simulation animation requiring
    extracellular spaces to be enabled by the current simulation
    configuration).

    Attributes
    ----------
    result : str
        Human-readable string justifying this failure. For generality, this
        string is neither capitalized *nor* punctuated.
    reason : str
        Human-readable string justifying this failure. For generality, this
        string is neither capitalized *nor* punctuated.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(self, result: str, reason: str) -> None:
        '''
        Initialize this exception.

        Parameters
        ----------
        result : str
            Human-readable string describing this failure. For generality, this
            string is expected to be neither capitalized *nor* punctuated.
        reason : str
            Human-readable string justifying this failure. For generality, this
            string is expected to be neither capitalized *nor* punctuated.
        '''

        # Capitalize all passed parameters.
        self.result = result
        self.reason = reason

        # Initialize our superclass, concatenating these strings into a single
        # human-readable exception message.
        super().__init__('{}: {}.'.format(self.result, self.reason))



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level shell-specific directory facilities.
'''

# ....................{ IMPORTS                            }....................
import os
from beartype.typing import Iterator
from betse.util.io.log import logs
from contextlib import contextmanager

# ....................{ GETTERS                            }....................
def get_cwd_dirname() -> str:
    '''
    **Current working dirname** (i.e., absolute path of the current working
    directory (CWD)) of the active Python process.

    Unless subsequently changed, this is the absolute path of the directory
    from which this application was initially run.
    '''

    return os.getcwd()

# ....................{ SETTERS                            }....................
def set_cwd(dirname: str) -> None:
    '''
    Set the **current working directory** (CWD) of the active Python process to
    the passed directory.

    This function permanently changes the CWD for the remainder of this
    process. For a robust alternative changing the CWD for a single code block,
    consider using the :func:`setting_cwd` context manager instead.

    Parameters
    -----------
    dirname : str
        Absolute or relative path of the directory to change to.
    '''

    # Log this change.
    logs.log_debug('Changing current working directory to: %s', dirname)

    # Change to this directory.
    os.chdir(dirname)

# ....................{ CONTEXTS                           }....................
@contextmanager
def setting_cwd(dirname: str) -> Iterator:
    '''
    Context manager setting the **current working directory** (CWD) of the
    active Python process to the passed directory for the duration of this
    context.

    This context manager guaranteeably reverts the CWD to the prior CWD even if
    an exception is raised (e.g., due to a missing directory).

    Parameters
    -----------
    dirname : str
        Absolute or relative dirname of the directory to change to.

    Returns
    -----------
    contextlib._GeneratorContextManager
        Context manager changing the CWD as described above.

    Yields
    -----------
    None
        Since this context manager yields no values, the caller's ``with``
        statement must be suffixed by *no* ``as`` clause.

    See Also
    -----------
    https://stackoverflow.com/a/24176022/2809027
        StackOverflow answer strongly inspiring this implementation.

    Examples
    -----------
        >>> from betse.util.paths import dirs
        >>> print('CWD: ' + dirs.get_cwd_dirname())
        CWD: /home/azrael
        >>> with dirs.current('/home/uriel/urial/nuriel/uryan/jeremiel'):
        ...     print('CWD: ' + dirs.get_cwd_dirname())
        ...     raise ValueError(
        ...         'But unknown, abstracted, brooding secret the dark power hid')
        CWD: /home/uriel/urial/nuriel/uryan/jeremiel
        ValueError: But unknown, abstracted, brooding secret the dark power hid.
        >>> print('CWD: ' + dirs.get_cwd_dirname())
        CWD: /home/azrael
    '''

    # Absolute path of the current CWD.
    dirname_prior = get_cwd_dirname()

    # Temporarily change to the passed directory. Since Python performs this
    # change only if this call raises no exceptions, this call need *NOT* be
    # embedded in the "try" block below.
    set_cwd(dirname)

    # Yield control to the body of the caller's "with" block.
    try:
        yield
    # Revert to the prior CWD even if that block raised an exception.
    finally:
        os.chdir(dirname_prior)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level pathname (e.g., basename, dirname, filetype) facilities.
'''

# ....................{ IMPORTS                            }....................
import errno, os
from betse.exceptions import BetsePathnameException
from betse.util.io.log import logs
from betse.util.type.decorator.decmemo import func_cached
from betse.util.type.types import type_check, ContainerTypes, StrOrNoneTypes
from os import path as os_path

# ....................{ CONSTANTS                          }....................
INVALID_PATHNAME = '\0'
'''
Pathname guaranteed to be invalid on all supported platforms.

This pathname consists of a single null byte. The operating system kernels for
both Windows _and_ all POSIX-compatible platforms (e.g., Linux, OS X) prohibit
pathnames containing null bytes by returning non-zero failure codes from
filesystem-centric kernel functions passed such pathnames. Python translates
these errors into raised exceptions. Under Linux, for example, a
:class:`TypeError` exception of message ``"embedded NUL character"`` is raised.
'''

# ....................{ EXCEPTIONS ~ absolute              }....................
def die_if_absolute(*pathnames: str) -> None:
    '''
    Raise an exception if any passed pathname is absolute.

    Equivalently, this function raises an exception unless unless all passed
    pathnames are relative.

    Raises
    ----------
    BetsePathnameException
        If any such pathname is absolute.

    See Also
    ----------
    :func:`is_absolute`
        Further details.
    '''

    for pathname in pathnames:
        if is_absolute(pathname):
            raise BetsePathnameException(
                'Pathname "{}" absolute rather than relative.'.format(
                    pathname))


def die_if_relative(*pathnames: str) -> None:
    '''
    Raise an exception if any passed pathname is relative.

    Equivalently, this function raises an exception unless unless all passed
    pathnames are absolute.

    Raises
    ----------
    BetsePathnameException
        If any such pathname is relative.

    See Also
    ----------
    :func:`is_relative`
        Further details.
    '''

    for pathname in pathnames:
        if is_relative(pathname):
            raise BetsePathnameException(
                'Pathname "{}" relative rather than absolute.'.format(
                    pathname))

# ....................{ EXCEPTIONS ~ basename              }....................
def die_if_basename(pathname: str) -> None:
    '''
    Raise an exception if the passed pathname is a **pure basename** (i.e.,
    contains no directory separators).

    See Also
    ----------
    :func:`is_basename`
        Further details.
    '''

    if is_basename(pathname):
        raise BetsePathnameException(
            'Pathname "{}" contains no directory separators '
            "(i.e., '{}' characters).".format(pathname, os_path.sep))


def die_unless_basename(pathname: str) -> None:
    '''
    Raise an exception unless the passed pathname is a **pure basename** (i.e.,
    contains one or more directory separators).

    See Also
    ----------
    :func:`is_basename`
        Further details.
    '''

    if not is_basename(pathname):
        raise BetsePathnameException(
            'Pathname "{}" contains one or more directory separators'
            "(i.e., '{}' characters).".format(pathname, os_path.sep))

# ....................{ EXCEPTIONS ~ parent                }....................
def die_unless_parent(parent_dirname: str, child_pathname: str) -> bool:
    '''
    Raise an exception unless the second passed pathname is a child of (i.e.,
    prefixed by) the first passed dirname.

    Parameters
    ----------
    parent_dirname : str
        Absolute or relative pathname of the candidate parent directory.
    child_pathname : str
        Absolute or relative pathname of the candidate child path.

    See Also
    ----------
    :func:`is_parent`
        Further details.
    '''

    if not is_parent(parent_dirname, child_pathname):
        raise BetsePathnameException(
            'Pathname "{}" not in dirname "{}".'.format(
                child_pathname, parent_dirname,))

# ....................{ EXCEPTIONS ~ filetype              }....................
def die_unless_filetype_equals(pathname: str, filetype: str) -> None:
    '''
    Raise an exception unless the passed path has the passed filetype.

    See Also
    ----------
    :func:`is_filetype_equals`
        Further details.
    '''

    if not is_filetype_equals(pathname, filetype):
        raise BetsePathnameException(
            'Pathname "{}" filetype not "{}".'.format(pathname, filetype))

# ....................{ TESTERS                            }....................
@type_check
def is_pathname(pathname: str) -> bool:
    '''
    ``True`` only if the passed string is a valid pathname (either absolute or
    relative) for the root filesystem of the current platform.

    Under:

    * POSIX-compatible OSes:

      * Valid pathnames are non-empty strings containing:

        * No null byte characters.
        * No ``/``-delimited path component longer than 255 characters.

      * The root filesystem is the filesystem mounted to the root directory
        ``/``.

    * Microsoft OSes:

      * Valid pathnames are non-empty strings satisfying `various constraints
        <https://msdn.microsoft.com/en-us/library/windows/desktop/aa365247%28v=vs.85%29.aspx>`_
        too numerous to document here.
      * The root filesystem is the filesystem to which this instance of Windows
        was installed, also given by the ``%HOMEDRIVE%`` environment variable.
    '''

    # Avoid circular import dependencies.
    from betse.util.os.brand import windows
    from betse.util.os.brand.windows import WindowsErrorType

    # If this is the empty string, it cannot by definition be a valid pathname.
    if not pathname:
        return False

    # The only cross-platform and -filesystem portable means of validating
    # pathnames is to parse exceptions raised by the kernel-dependent os.stat()
    # or os.lstat() functions for metadata indicating invalid pathname syntax.
    # All other approaches (e.g., regex string matching) fail for common edge
    # cases. See also:
    #     https://stackoverflow.com/a/34102855/2809027
    try:
        # Strip this pathname's Windows drive specifier (e.g., "C:\") if any.
        # Since Windows prohibits path components from containing ":"
        # characters, failing to strip this ":"-suffixed prefix would
        # erroneously invalidate all valid absolute Windows pathnames.
        _, pathname = os_path.splitdrive(pathname)

        # Absolute path of a directory guaranteed to exist.
        #
        # To avoid race conditions with external processes concurrently
        # modifying the filesystem, the passed pathname cannot be tested as is.
        # Only path components split from this pathname are safely testable.
        # Why? Because os.stat() and os.lstat() raise "FileNotFoundError"
        # exceptions when passed pathnames residing in non-existing directories
        # regardless of whether these pathnames are invalid or not. Directory
        # existence takes precedence over pathname invalidity. Hence, the only
        # means of testing whether pathnames are invalid or not is to:
        #
        # 1. Split the passed pathname into path components (e.g.,
        #    "/foo/bar" into "['', 'foo', 'bar']").
        # 2. For each path component:
        #    1. Join the pathname of a directory guaranteed to exist and the
        #       current path component into a new pathname (e.g., "/bar").
        #    2. Pass that pathname to os.stat() or os.lstat(). If that
        #       pathname and hence current path component is invalid, this
        #       call is guaranteed to raise an exception exposing the type
        #       of invalidity rather than a generic "FileNotFoundError"
        #       exception. Why? Because that pathname resides in an
        #       existing directory. Circular logic is circular.
        #
        # Is a directory guaranteed to exist? Yes, but typically only one: the
        # root directory for the current filesystem. Passing pathnames residing
        # in any other directories to os.stat() or os.lstat() invites
        # mind-flaying race conditions, even for directories previously tested
        # to exist. External processes cannot be prevented from concurrently
        # removing those directories after those tests have been performed but
        # before those pathnames are passed to os.stat() or os.lstat().
        #
        # Did we mention this should be shipped with Python already?
        root_dirname = get_root_dirname()
        assert os_path.isdir(root_dirname)   # ...Murphy and her dumb Law

        # Test whether each path component split from this pathname is valid
        # or not. Most path components will *NOT* actually physically exist.
        for pathname_part in pathname.split(os_path.sep):
            try:
                os.lstat(root_dirname + pathname_part)
            # If an OS-specific exception is raised, its error code indicates
            # whether this pathname is valid or not. Unless this is the case,
            # this exception implies an ignorable kernel or filesystem
            # complaint (e.g., path not found or inaccessible).
            #
            # Only the following exceptions indicate invalid pathnames:
            #
            # * Instances of the Windows-specific "WindowsError" class
            #   defining the "winerror" attribute whose value is
            #   "ERROR_INVALID_NAME". Under Windows, "winerror" is more
            #   fine-grained and hence useful than the generic "errno"
            #   attribute. When a too-long pathname is passed, for example,
            #   "errno" is "ENOENT" (i.e., no such file or directory) rather
            #   than "ENAMETOOLONG" (i.e., file name too long).
            # * Instances of the cross-platform "OSError" class defining the
            #   generic "errno" attribute whose value is either:
            #   * Under most POSIX-compatible OSes, "ENAMETOOLONG".
            #   * Under some edge-case OSes (e.g., SunOS, *BSD), "ERANGE".
            except OSError as exc:
                # True only if this pathname is invalid (as detailed above).
                is_pathname_invalid = (
                    windows.is_exception_pathname_invalid(exc)
                    if isinstance(exc, WindowsErrorType)
                    else exc.errno in {errno.ENAMETOOLONG, errno.ERANGE})

                # If this pathname is invalid, log a warning and return False.
                if is_pathname_invalid:
                    logs.log_warning(
                        'Pathname "{}" invalid: {}'.format(
                            pathname, exc.strerror))
                    return False
    # If a "TypeError" exception was raised, it almost certainly has the
    # error message "embedded NUL character" indicating an invalid pathname.
    except TypeError as exc:
        logs.log_warning(
            'Pathname "{}" invalid: {}'.format(pathname, str(exc)))
        return False
    # If no exception was raised, all path components and hence this
    # pathname itself are valid. (Praise be to the curmudgeonly python.)
    else:
        return True
    # If any other exception was raised, this is an unrelated fatal issue
    # (e.g., a bug). Permit this exception to unwind the call stack.

# ....................{ TESTERS ~ absolute                 }....................
@type_check
def is_absolute(pathname: str) -> bool:
    '''
    ``True`` only if the passed path is absolute.

    The definition of "absolute" depends on the current operating system.
    Specifically, under:

    * POSIX-compatible systems, absolute paths are prefixed by ``/``.
    * Microsoft Windows, absolute paths are prefixed by an optional drive
      indicator (e.g., ``C:``) followed by ``\\``.
    '''

    return os_path.isabs(pathname)


def is_relative(pathname: str) -> bool:
    '''
    ``True`` only if the passed path is relative.

    The definition of "relative" depends on the current operating system.
    Specifically, under:

    * POSIX-compatible systems, relative paths are *not* prefixed by ``/``.
    * Microsoft Windows, relative paths are *not* prefixed by an optional drive
      indicator (e.g., ``C:``) followed by ``\\``.
    '''

    return not is_absolute(pathname)

# ....................{ TESTERS ~ basename                 }....................
@type_check
def is_basename(pathname: str) -> bool:
    '''
    ``True`` only if the passed pathname is a **pure basename** (i.e., contains
    no directory separators and hence no directory components).
    '''

    return os_path.sep not in pathname

# ....................{ TESTERS ~ parent                   }....................
@type_check
def is_parent(parent_dirname: str, child_pathname: str) -> bool:
    '''
    ``True`` only if the second passed pathname is a child of (i.e., prefixed
    by) the first passed dirname.

    Parameters
    ----------
    parent_dirname : str
        Absolute or relative pathname of the candidate parent directory.
    child_pathname : str
        Absolute or relative pathname of the candidate child path.

    Returns
    ----------
    bool
        ``True`` only if this child path is a child of this parent directory.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.text.string import strs

    # Absolute or relative pathname of the candidate parent directory, suffixed
    # by this platform's directory separator to ensure this child is actually a
    # contained child of this parent directory.
    parent_dirname_suffixed = strs.add_suffix_unless_found(
        text=parent_dirname, suffix=os_path.sep)

    # Return true only if this child pathname is suffixed by this dirname.
    return strs.is_prefix(text=child_pathname, prefix=parent_dirname_suffixed)

# ....................{ TESTERS ~ filetype                 }....................
@type_check
def is_filetype(pathname: str) -> bool:
    '''
    ``True`` only if the passed pathname is suffixed by a filetype (i.e.,
    contains one or more ``.`` delimiters).

    Parameters
    ----------
    pathname : str
        Pathname to be tested.

    Returns
    ----------
    bool
        ``True`` only if this pathname is suffixed by a filetype.
    '''

    # Return true only if this pathname contains one or more "." delimiters.
    return '.' in pathname


@type_check
def is_filetype_equals(pathname: str, filetype: str) -> bool:
    '''
    ``True`` only if the passed pathname has a filetype *and* the **last
    filetype** (i.e., last ``.``-prefixed substring of the basename) of this
    pathname is the passed filetype.

    The passed filetype may contain arbitrarily many ``.`` characters,
    including an optional prefixing ``.``. In all cases, this function behaves
    as expected.

    Parameters
    ----------
    pathname : str
        Pathname to be tested.
    filetype : str
        Filetype to test whether this pathname is suffixed by.

    Returns
    ----------
    bool
        ``True`` only if this pathname has this filetype.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.text.string import strs

    # Prefix this filetype by "." if needed.
    filetype = dot_filetype(filetype)

    # Return true only if this pathname is suffixed by this filetype.
    return strs.is_suffix(pathname, filetype)


@type_check
def is_filetype_undotted_in(
    pathname: str, filetypes_undotted: ContainerTypes) -> bool:
    '''
    ``True`` only if the passed pathname has a filetype *and* the **last
    undotted filetype** (i.e., last ``.``-prefixed substring of the basename
    excluding this ``.``) of this pathname is an item of the passed container.

    Parameters
    ----------
    pathname : str
        Pathname to be tested.
    filetypes: ContainerTypes
        Container of all undotted filetypes to test this pathname against. Each
        item of this container *must* be a string not prefixed by the ``.``
        character.

    Returns
    ----------
    bool
        ``True`` only if this pathname has a filetype *and* the last filetype
        of this pathname is an item of this container.
    '''

    # Undotted filetype of this filename if any *OR* "None" otherwise.
    filetype_undotted = get_filetype_undotted_or_none(pathname)

    # Return true only if...
    return (
        # This pathname is suffixed by a filetype.
        filetype_undotted is not None and
        # This filetype is in this container.
        filetype_undotted in filetypes_undotted
    )

# ....................{ GETTERS                            }....................
@type_check
def get_basename(pathname: str) -> str:
    '''
    **Basename** (i.e., last component) of the passed path.
    '''

    return os_path.basename(pathname)

# ....................{ GETTERS ~ dirname                  }....................
def get_dirname(pathname: str) -> str:
    '''
    **Dirname** (i.e., parent directory) of the passed path if this path has a
    dirname *or* raise an exception otherwise.
    '''

    # Raise an exception unless this path has a dirname.
    die_if_basename(pathname)

    # This dirname.
    dirname = get_dirname_or_empty(pathname)

    # Assert this dirname's non-emptiness. Technically, the above call *SHOULD*
    # have ensured this. This is a sufficiently critical function, however, to
    # warrant asserting this fact.
    assert len(dirname), 'Pathname "{}" dirname empty.'.format(pathname)
    return dirname


def get_dirname_or_cwd(pathname: str) -> str:
    '''
    **Dirname** (i.e., parent directory) of the passed path if this path has a
    dirname *or* the current working directory otherwise.
    '''

    # Avoid circular import dependencies.
    from betse.util.os.shell import shelldir
    dirname = get_dirname_or_empty(pathname)
    return dirname if dirname else shelldir.get_cwd_dirname()


@type_check
def get_dirname_or_empty(pathname: str) -> str:
    '''
    **Dirname** (i.e., parent directory) of the passed path if this path has a
    dirname *or* the empty string otherwise.
    '''

    return os_path.dirname(pathname)

# ....................{ GETTERS ~ dirname : system         }....................
@func_cached
def get_root_dirname() -> str:
    '''
    Absolute dirname of the root directory suffixed by a directory separator.

    The definition of "root directory" conditionally depends on the current
    platform. Specifically, if this platform is:

    * POSIX-compatible (e.g., Linux, OS X), this is simply ``/``.
    * Microsoft Windows, this is the value of the ``%HOMEDRIVE%`` environment
      variable. This is the ``:``-suffixed letter of the drive to which Windows
      was originally installed -- typically but *not* necessarily ``C:\\``.
    '''

    # Avoid circular import dependencies.
    from betse.util.os.brand import windows
    from betse.util.os.shell import shellenv

    # Return this dirname.
    return (
        shellenv.get_var_or_default('HOMEDRIVE', 'C:') + os_path.sep
        if windows.is_windows_vanilla() else
        os_path.sep
    )


@func_cached
def get_home_dirname() -> str:
    '''
    Absolute dirname of the home directory of the current user if found *or*
    raise an exception otherwise (i.e., if this directory is *not* found).

    Raises
    ----------
    BetseDirException
        If this user has no home directory.
    '''

    # Avoid circular import dependencies.
    from betse.util.path import dirs

    # Absolute path of this directory.
    home_dirname = canonicalize('~')

    # If this directory is not found, fail.
    dirs.die_unless_dir(home_dirname)

    # Return this directory's path.
    return home_dirname

# ....................{ GETTERS ~ filetype                 }....................
@type_check
def get_pathname_sans_filetype(pathname: str) -> str:
    '''
    Passed pathname without the last ``.``-prefixed filetype (including this
    prefix) if present *or* this pathname as is otherwise.
    '''

    return os_path.splitext(pathname)[0]


@type_check
def get_pathname_sans_filetypes(pathname: str) -> str:
    '''
    Passed pathname with all suffixing ``.``-prefixed filetypes (including each
    such prefix) if present removed *or* this pathname as is otherwise.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.text.string import strs

    # Dirname of this path if any or the empty string otherwise.
    dirname = get_dirname_or_empty(pathname)

    # Basename of this path.
    basename = get_basename(pathname)

    # Strip all characters following the first "." from this basename.
    basename = strs.remove_suffix_prefixed(basename, '.')

    # If this path contains a dirname, return the concatenation of this dirname
    # and stripped basename; else only return this stripped basename.
    return join(dirname, basename) if dirname else basename

# ....................{ GETTERS ~ filetype : undotted      }....................
@type_check
def get_filetype_dotted_or_none(pathname: str) -> StrOrNoneTypes:
    '''
    ``.``-prefixed **last filetype** (i.e., last ``.``-prefixed substring of
    the basename) of the passed pathname if this pathname is suffixed by a
    filetype *or* ``None`` otherwise.

    If this pathname is suffixed by two or more filetypes (e.g.,
    ``odium.reigns.tar.gz``), only the last such filetype is returned.
    '''

    # "."-prefixed filetype of this pathname.
    filetype = os_path.splitext(pathname)[1]

    # Return this string as is if non-empty or "None" otherwise.
    return filetype or None


def get_filetype_undotted(pathname: str) -> str:
    '''
    **Last filetype** (i.e., last ``.``-prefixed substring of the basename
    excluding this ``.``) of the passed pathname.

    Raises
    ----------
    :exc:`BetsePathnameException`
        If this pathname is suffixed by no filetype (i.e., if this pathname
        contains no ``.`` character).

    See Also
    ----------
    :func:`get_filetype_dotted_or_none`
        Further details.
    '''

    # Filetype of this pathname if any or "None" otherwise.
    filetype = get_filetype_undotted_or_none(pathname)

    # If this pathname has no filetype, raise an exception.
    if filetype is None:
        raise BetsePathnameException(
            'Path "{}" has no filetype.'.format(pathname))

    # Else, return this filetype.
    return filetype


@type_check
def get_filetype_undotted_or_none(pathname: str) -> StrOrNoneTypes:
    '''
    **Last filetype** (i.e., last ``.``-prefixed substring of the basename
    excluding this ``.``) of the passed pathname if this pathname is suffixed
    by a filetype *or* ``None`` otherwise.

    See Also
    ----------
    :func:`get_filetype_dotted_or_none`
        Further details.
    '''

    # "."-prefixed filetype of this pathname.
    filetype = get_filetype_dotted_or_none(pathname)

    # Return this string stripped of this prefix if non-empty or "None"
    # otherwise.
    return filetype[1:] if filetype is not None else None

# ....................{ DOTTERS                            }....................
@type_check
def dot_filetype(filetype: str) -> str:
    '''
    The passed filetype prefixed by a ``.`` character if needed (i.e., if not
    already prefixed by this character *or* this filetype as is otherwise).

    Parameters
    ----------
    filetype : str
        Filetype to be prefixed.

    Returns
    ----------
    str
        This filetype prefixed by a ``.`` character if *not* already prefixed.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.text.string import strs

    # Return this filetype prefixed by "." if needed.
    return strs.add_prefix_unless_found(text=filetype, prefix='.')


@type_check
def undot_filetype(filetype: str) -> str:
    '''
    The passed filetype prefixed by no ``.`` character.

    Parameters
    ----------
    filetype : str
        Filetype to be unprefixed.

    Returns
    ----------
    str
        Either:

        * If this filetype is prefixed by a ``.`` character, this filetype
          stripped of this prefix.
        * Else, this filetype unmodified.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.text.string import strs

    # Return this filetype prefixed by "." if needed.
    return strs.remove_prefix_if_found(text=filetype, prefix='.')

# ....................{ CANONICALIZERS                     }....................
@type_check
def join_and_canonicalize(*partnames: str) -> str:
    '''
    Canonical form of the path produced by joining (i.e., concatenating) the
    passed pathnames with the directory separator specific to the current
    platform.

    This higher-level function is a convenience wrapper encapsulating both the
    lower-level :func:`join` and :func:`canonicalize` functions.
    '''

    return canonicalize(join(*partnames))


@type_check
def canonicalize(pathname: str) -> str:
    '''
    **Canonical form** (i.e., unique absolute pathname *after* transitively
    resolving all symbolic links) of the passed path.

    Specifically, this function (in order):

    #. Transitively resolves all symbolic links, producing a pathname that
       #either:

       * Does not exist.
       * Does exist but is *not* a symbolic link.

    #. Performs **tilde expansion,** replacing a ``~`` character prefixing this
       path by the absolute path of the current user's home directory.

    #. Performs **path normalization,** which (in no particular order):

      * Collapses redundant separators (e.g., converting ``//`` to ``/``).
      * Converts explicit relative to absolute path components (e.g., converts
        ``../`` to the name of the parent directory of that component).
      * Converts implicit relative basenames to absolute pathnames (e.g.,
        converts ``sim_config.yaml`` to ``/tmp/sim_config.yaml`` when the
        current working directory is ``/tmp``).
    '''

    return os_path.realpath(os_path.expanduser(pathname))

# ....................{ JOINERS                            }....................
@type_check
def join(*partnames: str) -> str:
    '''
    Join (i.e., concatenate) the passed pathnames with the directory separator
    specific to the current platform.

    Caveats
    ----------
    This high-level wrapper should *always* be called in lieu of the low-level
    :func:`os.path.join` function, which sadly violates Python's core "Explicit
    is better than implicit." design principle. Specifically, to quote official
    :func:`os.path.join` documentation:

        If a component is an absolute path, all previous components are thrown
        away and joining continues from the absolute path component.

    Non-intuitive path mangling of that sort only promotes non-debuggable bugs.
    This function prevents these bugs by explicitly raising an exception if any
    passed pathname except the first is absolute.

    Parameters
    ----------
    partnames : tuple[str]
        Tuple of all pathnames to be joined. The first such pathname may be
        absolute or relative. All remaining pathnames *must* be relative.

    Returns
    ----------
    str
        Absolute or relative pathname joined from these pathnames.

    Raises
    ----------
    BetsePathnameException
        If any such pathname except the first is absolute.
    '''

    # If more than one partname was passed *AND* any such partname is absolute,
    # raise an exception. See above for commentary.
    if len(partnames) > 1:
        die_if_absolute(*partnames[1:])

    # Return the concatenation of these partnames.
    return os_path.join(*partnames)

# ....................{ RELATIVIZERS                       }....................
@type_check
def relativize(src_dirname: str, trg_pathname: str) -> str:
    '''
    Relative pathname of the second passed pathname relative to the first
    passed dirname.

    Parameters
    ----------
    src_dirname : str
        Absolute or relative pathname of the source directory.
    trg_pathname : str
        Absolute or relative pathname of the target path.

    Returns
    ----------
    str
        Absolute or relative pathname of the target path converted into a
        relative path relative to the source directory.
    '''

    return os_path.relpath(trg_pathname, start=src_dirname)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Unit tests exercising the :mod:`betse.util.type.iterable.iterables` submodule.
'''

# ....................{ IMPORTS                           }....................

# ....................{ TESTS ~ default                   }....................
#FIXME: Exercise all remaining edge cases... of which there are many. See the
#iterables.to_iterable() docstring for further details.
def test_to_iterable() -> None:
    '''
    Unit test the
    :func:`betse.util.type.iterable.iterables.to_iterable` function.
    '''

    # Defer heavyweight imports.
    from betse.util.type.iterable import iterables

    # Input tuple to be converted from.
    howl_tuple = (
        'incomparable blind streets of shuddering cloud',
        'and lightning in the mind leaping toward poles',
        'of Canada & Paterson,',
        'illuminating all the motionless world',
        'of Time between,',
    )

    # Input generator to be converted from.
    howl_generator = (stanza for stanza in howl_tuple)

    # Test whether this conversion preserves this input tuple as is.
    assert iterables.to_iterable(iterable=howl_tuple) is howl_tuple

    # Test whether this conversion converts this input tuple into an output
    # iterable of differing type.
    assert iterables.to_iterable(iterable=howl_tuple, cls=set) == set(
        howl_tuple)

    # Test whether this conversion converts this input generator into an output
    # tuple identical to this input tuple.
    assert iterables.to_iterable(iterable=howl_generator) == howl_tuple



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **non-string iterable sorting** (i.e., sorting of non-string objects
implementing the abstract base class :class:`collections.abc.Iterable`)
facilities.
'''

#FIXME: Generalize to support mappings. Doing so may prove mildly non-trivial,
#as unordered mapping types (e.g., plain "dict" instances) will need to be
#converted into ordered mapping types (e.g., "OrderedDict" instances).

# ....................{ IMPORTS                           }....................
from betse.util.type.types import type_check, IterableTypes
from operator import itemgetter

# ....................{ SORTERS ~ ascending               }....................
@type_check
def sort_ascending(iterable: IterableTypes) -> IterableTypes:
    '''
    Iterable sorted from the passed iterable in ascending order.

    Each item of this iterable is compared to each other item of this iterable
    under the ``<`` operator, thus implicitly calling the ``__le__()`` special
    method of each such item. Each item should ideally but *not* necessarily be
    of the same type. If each item is:

    * A string, these strings are sorted in **ascending lexicographic order**
      (i.e., traditional order of dead-tree dictionaries and encyclopedias).
    * A number (i.e., integer or a float), these numbers are sorted in
      **ascending numeric order.**

    Caveats
    ----------
    If the passed iterable is a generator, this function:

    * Internally coerces this generator into a list *before* attempting to
      sort the items yielded by this generator.
    * Externally returns a new list rather than a new generator.

    This constraint is mandated by the :func:`sorted` builtin, which raises the
    following exception on receiving a generator:

       TypeError: cannot create 'generator' instances

    Parameters
    ----------
    iterable : IterableTypes
        Unsorted iterable to be returned sorted. For generality, this iterable
        is *not* modified by this function.

    Returns
    ----------
    IterableTypes
        Iterable sorted from and of the same type as the passed iterable. For
        efficiency, this iterable is only a shallow rather than deep copy of
        the passed iterable. Note lastly that the class of the passed iterable
        *must* define an ``__init__()`` method accepting a list.
    '''

    # Defer to this lowor-level function. Since the default behaviour of the
    # sorted() builtin and thus this function is to sort in ascending
    # lexicographic order, no additional keyword arguments are required.
    return _sort_iterable(iterable)


@type_check
def sort_by_index_ascending(
    iterable: IterableTypes, subiterable_index: object) -> IterableTypes:
    '''
    Iterable of subiterables sorted from the passed iterable of subiterables in
    ascending order of the value of each element at the passed key or index of
    each subiterable of this iterable.

    Each item at the passed key or index of each subiterable of this iterable
    is compared to each other item at each other key or index of each other
    subiterable of this iterable via the ``<`` operator, implicitly calling the
    ``__le__()`` special method of these items. Each item is ideally but *not*
    necessarily of the same type. If each item is:

    * A string, these strings are sorted in **ascending lexicographic order**
      (i.e., traditional order of dead-tree dictionaries and encyclopedias).
    * A number (i.e., integer or a float), these numbers are sorted in
      **ascending numeric order.**

    Parameters
    ----------
    iterable : IterableTypes
        Unsorted iterable of subiterables to be returned sorted. For
        generality, neither this iterable nor these subiterables are modified
        by this function.
    subiterable_index : object
        Object with which to index each subiterable of this iterable. The type
        of this object *must* be a type accepted by the ``__getitem__()``
        special method of each subiterable. Specifically, if each subiterable
        is a:

        * **Mapping** (e.g., :class:`dict`), this object *must* be hashable.
        * **Sequence** (e.g., :class:`list`, :class:`tuple`), this object
          *must* be either:

          * An integer.
          * A :func:`slice` object.

    Returns
    ----------
    IterableTypes
        Iterable of subiterables sorted from and of the same type as the passed
        iterable of subiterables. For efficiency, this iterable is only a
        shallow rather than deep copy of the passed iterable. Note lastly that
        the class of the passed iterable *must* define an ``__init__()`` method
        accepting a list.

    See Also
    ----------
    :func:`sort_ascending`
        Further details.
    '''

    # Defer to this lowor-level function.
    #
    # For efficiency, items of each subiterable of this iterable are retrieved
    # via the function internally created and called by calling an instance of
    # the standard "itemgetter" class. While technically an instance of that
    # class, that class is intended to be treated as a simple function passed
    # the desired index returning a function passed a subiterable returning the
    # element at that index: e.g.,
    #
    #     >>> penguins = [('adelie', 0xFEEDFACE), ('gentoo', 0xDEADBEEF)]
    #     >>> penguin_id = itemgetter(2)
    #     >>> penguin_id(penguins[0]) == 0xFEEDFACE
    #     True
    #
    # Despite the internal complexity and object overhead imposed by the
    # "itemgetter" class, this approach has been definitively profiled to be
    # 126% faster on average than the traditional
    # "lambda subiterable: subiterable[subiterable_index]" approach --
    # presumably due to hidden overhead imposed by capturing all local
    # variables into the closure context encapsulated by the lambda. See also
    # the following Stackoverflow answer exhibiting this profiling:
    #
    #     https://stackoverflow.com/a/17243726/2809027
    return _sort_iterable(iterable, key=itemgetter(subiterable_index))

# ....................{ SORTERS ~ descending              }....................
@type_check
def sort_descending(iterable: IterableTypes) -> IterableTypes:
    '''
    Iterable sorted from the passed iterable in descending order.

    Each item of this iterable is compared to each other item of this iterable
    via the ``>`` operator, implicitly calling the ``__ge__()`` special method
    of these items. Each item is ideally but *not* necessarily of the same
    type. If each item is:

    * A string, these strings are sorted in **descending lexicographic order**
      (i.e., reverse order of dead-tree dictionaries and encyclopedias).
    * A number (i.e., integer or a float), these numbers are sorted in
      **descending numeric order.**

    See Also
    ----------
    :func:`sort_ascending`
        Further details.
    '''

    # Defer to this lowor-level function.
    return _sort_iterable(iterable, reverse=True)


@type_check
def sort_by_index_descending(
    iterable: IterableTypes, subiterable_index: object) -> IterableTypes:
    '''
    Iterable of subiterables sorted from the passed iterable of subiterables in
    descending order of the value of each element at the passed key or index of
    each subiterable of this iterable.

    Each element at the passed key or index of each subiterable of this
    iterable is compared to each other element at each other key or index of
    each other subiterable of this iterable via the ``>`` operator, implicitly
    calling the ``__ge__()`` special method of these elements. Each element is
    ideally but *not* necessarily of the same type. If each element is:

    * A string, these strings are sorted in **descending lexicographic order**
      (i.e., reverse order of dead-tree dictionaries and encyclopedias).
    * A number (i.e., integer or a float), these numbers are sorted in
      **descending numeric order.**

    See Also
    ----------
    :func:`sort_by_index_ascending`
        Further details.
    '''

    # Defer to this lowor-level function. See sort_by_index_descending() for
    # additional commentary, particularly for the "key" parameter.
    return _sort_iterable(
        iterable,
        key=itemgetter(subiterable_index),
        reverse=True,
    )

# ....................{ SORTERS ~ private                 }....................
@type_check
def _sort_iterable(iterable: IterableTypes, **kwargs) -> IterableTypes:
    '''
    Iterable sorted from the passed iterable according to the passed keyword
    arguments accepted by the :func:`sorted` builtin.

    Parameters
    ----------
    iterable : IterableTypes
        Unsorted iterable to be returned sorted. For generality, this iterable
        is *not* modified by this function.

    All remaining keyword arguments are passed as is to the :func:`sorted`
    builtin.

    Returns
    ----------
    IterableTypes
        Iterable sorted from and of the same type as the passed iterable. For
        efficiency, this iterable is only a shallow rather than deep copy of
        the passed iterable. Note lastly that the class of the passed iterable
        *must* define an ``__init__()`` method accepting a list.

    See Also
    ----------
    :func:`sort_ascending`
        Further details.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.iterable import generators

    # Input iterable to be sorted.
    #
    # If this iterable is a generator, coerce this generator into a list.
    # Generators are *NOT* safely sortable as is. See the function docsctring
    # for details. Note that coercing this generator into a list ensures that
    # the "list_sorted" generated below is efficiently returned as is without
    # requiring further inefficient type coercion (e.g., from a list to a
    # tuple).
    iterable_sortable = (
        list(iterable) if generators.is_generator(iterable) else iterable)

    # Type of both this iterable *AND* the output iterable to be returned.
    iterable_type = type(iterable_sortable)

    # Output list sorted from this iterable.
    list_sorted = sorted(iterable_sortable, **kwargs)

    # Return either:
    #
    # * If the input iterable is:
    #   * A list, efficiently return this output list as is, which already
    #     satisfies the exact type expected by the caller.
    # * Else, an iterable of the same type as that of the passed iterable.
    return list_sorted if iterable_type is list else iterable_type(list_sorted)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level vector subcache functionality.
'''

# ....................{ IMPORTS                            }....................
import numpy as np
from betse.exceptions import BetseSimVectorException
from betse.science.math import mathunit
from betse.science.math.cache.cacheabc import SimPhaseCacheABC
from betse.science.math.vector.veccls import VectorCellsCache
from betse.util.type.decorator.decmemo import property_cached

# ....................{ SUBCLASSES                         }....................
class SimPhaseCacheVectorCells(SimPhaseCacheABC):
    '''
    Simulation phase-specific vector subcache, persisting all previously
    constructed vectors for a single simulation phase.

    Attributes
    ----------
    layer : SimPhaseCacheLayerCellsVector
        Subcache of all vector-based layers constructed for this phase.
    '''

    # ..................{ INITIALIZORS                       }..................
    def __init__(self, *args, **kwargs) -> None:

        # Initialize our superclass with all passed parameters.
        super().__init__(*args, **kwargs)

        # Avoid circular import dependencies.
        from betse.science.math.cache.cachelyrvec import (
            SimPhaseCacheLayerCellsVector)

        # Classify all subcaches imported above.
        self.layer = SimPhaseCacheLayerCellsVector(self._phase)

    # ..................{ PROPERTIES ~ ions                  }..................
    #FIXME: Decorate to require that calcium ions be enabled.
    @property_cached
    def ion_calcium_intra(self) -> VectorCellsCache:
        '''
        Vector cache of all upscaled cellular calcium ion (Ca2+) concentrations
        over all sampled time steps of the current simulation phase, originally
        spatially situated at cell centres.
        '''

        return self._make_ion_intra_cache(ion_index=self._phase.sim.iCa)


    #FIXME: Decorate to require that hydrogen ions be enabled.
    @property_cached
    def ion_hydrogen_intra(self) -> VectorCellsCache:
        '''
        Vector cache of all logarithmically scaled cellular hydrogen ion (H+)
        concentrations over all sampled time steps of the current simulation
        phase, originally spatially situated at cell centres.

        For generality, these concentrations are logarathmically scaled to
        correspond exactly to cellular pH.
        '''

        # Create and return this cache, which logarithmically rather than
        # multiplicatively scales and hence *CANNOT* defer to the standardized
        # _make_ion_cache() method.
        return VectorCellsCache(
            phase=self._phase,
            times_cells_centre=[
                -np.log10(1.0e-3 * ions_concentration[self._phase.sim.iH])
                for ions_concentration in self._phase.sim.cc_time])

    # ..................{ PROPERTIES ~ voltage               }..................
    @property_cached
    def voltage_extra(self) -> VectorCellsCache:
        '''
        Vector cache of all upscaled **extracellular voltages** (i.e., voltages
        across all environmental grid spaces) over all sampled time steps of the
        current simulation phase, originally spatially situated at grid space
        centres.

        For readability of units in exported visuals (e.g., plots), voltages are
        cached upscaled from volts (V) to millivolts (mV).
        '''

        # Shape of the extracellular voltages array to be cached. While the
        # original "venv_time" is a two-dimensional array indexed first by
        # sampled time steps and then by flattened grid spaces, this is a
        # three-dimensional array indexed first by sampled time steps and then
        # by nonflattened grid space rows and columns as unique dimensions. Why?
        # Because layers plotting extracellular data assume the latter.
        voltage_extra_shape = (
            len(self._phase.sim.venv_time),) + self._phase.cells.X.shape

        # Create, return, and create this cache, both upscaled and reshaped as
        # detailed above.
        return VectorCellsCache(
            phase=self._phase,
            times_grids_centre=mathunit.upscale_units_milli(
                self._phase.sim.venv_time).reshape(voltage_extra_shape))


    @property_cached
    def voltage_membrane(self) -> VectorCellsCache:
        '''
        Vector cache of all upscaled **transmembrane voltages** (i.e., voltages
        across all gap junctions connecting intracellular membranes) over all
        sampled time steps of the current simulation phase, originally spatially
        situated at cell membrane midpoints.

        For readability of units in exported visuals (e.g., plots), voltages are
        cached upscaled from volts (V) to millivolts (mV).
        '''

        return VectorCellsCache(
            phase=self._phase,
            times_membranes_midpoint=mathunit.upscale_units_milli(
                self._phase.sim.vm_time))

    # ..................{ PRIVATE ~ ions                     }..................
    def _make_ion_intra_cache(self, ion_index: int) -> VectorCellsCache:
        '''
        Create and return a vector cache of all upscaled concentrations of the
        ion with the passed index over all sampled time steps of the current
        simulation phase, originally spatially situated at cell centres.

        Parameters
        ----------
        ion_index : int
            0-based index of the ion to cache concentrations for.

        Returns
        ----------
        VectorCellsCache
            Cache of all upscaled concentrations of this ion.

        Raises
        ----------
        BetseSimVectorException
            If no ion with this index exists (i.e., if this index is *not* in
            the range ``[0, len(self._phase.sim.cc_time[0]))``).
        '''

        # If no ion with this index exists, raise an exception.
        if not (0 <= ion_index < len(self._phase.sim.cc_time[0])):
            raise BetseSimVectorException(
                'Ion with index {} not found '
                '(i.e., not in range [0, {})).'.format(
                    ion_index, len(self._phase.sim.cc_time[0])))

        # Create and return this cache.
        return VectorCellsCache(
            phase=self._phase,
            times_cells_centre=mathunit.upscale_units_micro(
                ions_concentration[ion_index]
                for ions_concentration in self._phase.sim.cc_time))



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Microsoft Windows-specific facilities.

Caveats
----------
Operating system-specific logic is poor form and should be leveraged only where
necessary.
'''

#FIXME: Generalize this configuration to support PowerShell. Sadly, PowerShell
#is fundamentally insane. (Who didn't see that one coming?) In particular,
#PowerShell insanely interprets *ANY* attempt to write to stderr as a fatal
#exception. See our "appveyor.yml" configuration for further details.
#
#Since Microsoft clearly has no compelling interest in resolving this blatantly
#broken behaviour, we *MUST* do so on their behalf. To do so, consider:
#
#* Implementing a new betse.util.os.brand.windows.is_shell_powershell() tester
#  returning True only if the current shell environment is PowerShell. Note
#  that, as Windows only natively supports two shell environments, the related
#  is_shell_cmd() tester is trivially implementable as follows:
#    @func_cached
#    def is_shell_cmd() -> bool:
#        return not is_shell_powershell()
#* Generalize the _init_logger_root_handler_std() method defined below to:
#
#       # Expand this unconditional assignment...
#       self._logger_root_handler_stderr = StreamHandler(sys.stderr)
#
#       # ...into these conditional assignments.
#       if windows.is_windows() and windows.is_shell_powershell():
#           self._logger_root_handler_stderr = StreamHandler(sys.stdout)
#       else:
#           self._logger_root_handler_stderr = StreamHandler(sys.stderr)
#
#In optimistic theory, the above should suffice. </apathetic_shrug>
#FIXME: Actually, to implement this sanely, we'll need to generalize our use of
#"sys.stdout" and "sys.stderr" as follows:
#
#* Define a new betse.util.io.stderrs.get_stderr() function resembling:
#    @func_cached
#    def get_stderr() -> ????:
#       return (
#           sys.stdout
#           if windows.is_windows() and windows.is_shell_powershell():
#           sys.stderr)
#* Define a new betse.util.io.stdouts.get_stdout() function resembling:
#    def get_stdout() -> ????:
#        return sys.stdout
#* Replace all references to "sys.stdout" with calls to
#  betse.util.io.stdouts.get_stdout().
#* Replace all references to "sys.stderr" with calls to
#  betse.util.io.stderrs.get_stderr().
#FIXME: Unfortunately, implementing the requisite is_shell_powershell() tester
#in this submodule will prove highly non-trivial. The reason, of course, is
#that the Windows API provides no convenient means of differentiating a child
#PowerShell from CMD.exe process. That said, doing so is absolutely feasible.
#
#The standard solution to this issue is to leverage the third-party "psutil"
#package to query whether the name of the parent process matches that of a
#known PowerShell executable, as detailed here:
#    https://stackoverflow.com/a/55598796/2809027
#
#Since that package is C-based and hence heavyweight, however, the above
#solution should *ONLY* be pursued if everything written below fails. So, what
#alternatives do we happen? Simple! We implement is_shell_powershell() in pure
#Python using only standard packages and modules. But how do we do that? Here
#we go:
#
#* Invoke the "tasklist" executable effectively guaranteed to be available
#  in the current %PATH% in all modern (i.e., post-2012) Windows installations.
#* Pass that executable options that render the output amenable to trivial
#  regular expression-based parsing ala the official "talklist" documentation:
#    https://docs.microsoft.com/en-us/windows-server/administration/windows-commands/tasklist
#* Parse the "stdout" emitted by that executable using Python's standard "csv"
#  module ala this unofficial, clever gist:
#    https://gist.github.com/intco/6149781

# ....................{ IMPORTS                            }....................
import platform, sys
from betse.exceptions import BetseOSException
from betse.util.type.decorator.decmemo import func_cached
from betse.util.type.types import type_check, NoneType

# ....................{ TYPES                              }....................
WindowsErrorType = NoneType
'''
Windows-specific :class:`WindowsError` class if the current platform is Windows
*or* the :class:`NoneType` class otherwise.

Since all exceptions subclass the root :class:`Exception` superclass *and*
since the :class:`NoneType` class does not do so, this class is only an
exception class under Windows. In particular, attempting to either catch this
class as an exception *or* to test whether a caught exception is an instance of
this class is guaranteed to both be safe and behave as expected regardless of
the current platform. In short: don't worry, be emoji.
'''

# For use in type checking, this class is safely defined at the top-level in a
# platform-agnostic manner.
try:
    WindowsErrorType = WindowsError
except:
    pass

# ....................{ CONSTANTS ~ error codes            }....................
# For conformance, the names of all error code constants defined below are
# exactly as specified by Microsoft itself. Sadly, Python fails to provide
# these magic numbers for us.

_ERROR_INVALID_NAME = 123
'''
Microsoft Windows-specific error code indicating an invalid pathname.

See Also
----------
https://msdn.microsoft.com/en-us/library/windows/desktop/ms681382%28v=vs.85%29.aspx
    Official listing of all such codes.
'''

# ....................{ EXCEPTIONS                         }....................
def die_unless_windows() -> None:
    '''
    Raise an exception unless the current platform is Microsoft Windows.

    See Also
    ----------
    :func:`is_windows`
        Further details.
    '''

    # Avoid circular import dependencies.
    from betse.util.os import oses

    # If the current platform is *NOT* Windows, raise an exception.
    if not is_windows():
        raise BetseOSException(
            'Current platform {} not Windows.'.format(oses.get_name()))

# ....................{ TESTERS                            }....................
@func_cached
def is_windows() -> bool:
    '''
    ``True`` only if the current platform is Microsoft Windows.

    This function reports ``True`` for both vanilla and Cygwin Microsoft
    Windows (both of which commonly require special Windows-specific handling)
    but *not* the Windows Subsystem for Linux (WSL) (which accurately
    masquerades as Linux and hence does *not* commonly require special
    Windows-specific handling).
    '''

    return is_windows_vanilla() or is_windows_cygwin()


@func_cached
def is_windows_cygwin() -> bool:
    '''
    ``True`` only if the current platform is **Cygwin Microsoft Windows**
    (i.e., running the Cygwin POSIX compatibility layer).
    '''

    return sys.platform == 'cygwin'


@func_cached
def is_windows_vanilla() -> bool:
    '''
    ``True`` only if the current platform is **vanilla Microsoft Windows**
    (i.e., *not* running the Cygwin POSIX compatibility layer).
    '''

    return sys.platform == 'win32'


@func_cached
def is_windows_wsl() -> bool:
    '''
    ``True`` only if the current platform is **Windows Subsystem for Linux
    (WSL)** (i.e., the Microsoft-flavoured Linux kernel optionally supported by
    Windows 10).

    See Also
    ----------
    https://www.reddit.com/r/bashonubuntuonwindows/comments/85jghk/how_to_allow_python_to_know_im_on_windows/dvxwy9t
        Reddit post strongly inspiring this implementation.
    '''

    # Avoid circular import dependencies.
    from betse.util.os.brand import linux

    # If the active Python interpreter is *NOT* operating under a Linux kernel,
    # return false immediately.
    if not linux.is_linux():
        return False
    # Else, this interpreter is operating under a Linux kernel.

    # Flavour of this Linux kernel.
    kernel_flavour = platform.uname()[3]

    # Return true only if this is a Microsoft-flavoured Linux kernel.
    return 'microsoft' in kernel_flavour

# ....................{ TESTERS ~ path                     }....................
@type_check
def is_exception_pathname_invalid(exception: WindowsErrorType) -> bool:
    '''
    ``True`` only if the passed Windows-specific exception was raised from an
    erroneous attempt to read or write an invalid pathname.
    '''

    # If the current platform is *NOT* Windows, raise an exception.
    die_unless_windows()

    # The above type checking guarantees this exception to be an instance
    # of the Windows-specific "WindowsError" class. Test the "winerror"
    # attribute exclusive to such intsances.
    return exception.winerror == _ERROR_INVALID_NAME

# ....................{ TESTERS ~ version                  }....................
@func_cached
def is_version_10_or_newer() -> bool:
    '''
    ``True`` only if the current platform is Windows >= 10 (i.e., either
    Windows 10 or a newer version of Windows).
    '''

    # Avoid circular import dependencies.
    from betse.util.os.kernels import is_version_greater_than_or_equal_to

    # Return true only if this platform is Windows *AND* this version of the
    # Windows kernel is at least 10.
    return is_windows() and is_version_greater_than_or_equal_to('10.0.0')

# ....................{ GETTERS                            }....................
@func_cached
def get_api_version() -> str:
    '''
    Human-readable ``.``-delimited version specifier of the Windows API
    (WinAPI, Win32) underlying the current Windows installation (e.g.,
    ``6.2.9200``).
    '''

    # If the current platform is *NOT* Windows, raise an exception.
    die_unless_windows()

    # Return the second item of the 4-tuple "(release, version, csd, ptype)"
    # returned by this function.
    return platform.win32_ver()[1]



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Matplotlib-specific **colormap** (i.e., objects mapping from low-level
scientific data values to high-level RGBA colour values displaying that data)
facilities.

See Also
----------
https://matplotlib.org/examples/color/colormaps_reference.html
    Canonical colormap example visually depicting all default colormaps bundled
    with matplotlib.
'''

# ....................{ IMPORTS                            }....................
import numpy as np
from betse.exceptions import BetseSequenceException
from betse.util.io.log import logs
from betse.util.type.iterable import sequences
from betse.util.type.numeric import ints
from betse.util.type.types import type_check, SequenceTypes
from matplotlib import colormaps
from matplotlib.colors import (
    Colormap,
    LinearSegmentedColormap,
)

# ....................{ GETTERS                            }....................
@type_check
def get_colormap(name: str) -> Colormap:
    '''
    Matplotlib colormap with the passed name, including both standard colormaps
    bundled with matplotlib *and* application-specific colormaps registered by
    this submodule.

    This function is a convenience wrapper for the
    :meth:`matplotlib.colormaps.__getindex__` dunder method, provided only as a
    human-readable utility to assist readability and maintainability.

    Parameters
    ----------
    name : str
        Name of the colormap to be retrieved. If this is:

        * A standard colormap bundled with matplotlib, this should be the name
          of the attribute in the :mod:`matplotlib.cm` module corresponding to
          the desired colormap (e.g., ``Blues``, ``jet``, ``rainbow``).
        * An application-specific colormap registered by this submodule, this
          should be the ``name`` parameter initializing an
          :class:`_MplColormapScheme` instance contained in the
          ``COLORMAP_SCHEMES`` tuple defined within the :func:`init` function.

    Returns
    ----------
    Colormap
        Matplotlib colormap with this name.

    See Also
    ----------
    http://matplotlib.org/examples/color/colormaps_reference.html
        List of supported colormaps.
    '''

    return colormaps[name]

# ....................{ ITERATORS                          }....................
def iter_colormap_names() -> SequenceTypes:
    '''
    Sequence of the names of all colormaps currently registered with matplotlib
    (in sorted lexicographic order).

    Caveats
    -----------
    **This function should be called only after this submodule has been
    initialized** (i.e., after the :func:`init` function has been called).
    While this function *is* safely callable beforehand, doing so prevents this
    function from returning the names of all application-specific colormaps
    registered by that function.

    **This function does not memoize this sequence** (e.g., by decorating this
    function with :func:`betse.util.type.decorator.decmemo.func_cached`). This
    sequence may change over the lifetime of this application and hence
    *cannot* be implicitly cached on the first call to this function. Instead,
    callers may prefer to manually cache this sequence for efficiency.
    '''

    # In theory, this implementation would simply defer to the existing
    # matplotlib.pyplot.colormaps() function returning the exact same sequence.
    # In practice, doing so is infeasible in the general case. Why? Because
    # merely importing the "matplotlib.pyplot" submodule incurs non-trivial,
    # non-deterministic, and irreversible side effects -- notably, the
    # selection of a default backend.
    #
    # Inevitably, some external caller will call this function *BEFORE*
    # selecting a default backend, implying that this function *CANNOT* import
    # from the "matplotlib.pyplot" submodule. Instead, this function inlines the
    # body of the matplotlib.pyplot.colormaps() function here. Fortunately, the
    # body of this function reduces to a trivial one-liner.
    return sorted(colormaps)

# ....................{ INITIALIZERS                       }....................
def init() -> None:
    '''
    Initialize this module by registering all application-specific colormaps
    with matplotlib, enabling these colormaps to be trivially retrieved with
    the standard :func:`matplotlib.cm.get_cmap` function.

    This function is intended to be called at matplotlib initialization time.
    '''

    # Enable these globals to be locally assigned to below.
    global _is_initted

    # ..................{ INITTED                            }..................
    # If this submodule has already been initialized, reduce to a noop.
    if _is_initted:
        logs.log_debug(
            'Ignoring attempt to reregister custom matplotlib colormaps...')
        return
    # Else, this submodule has *NOT* yet been initialized.

    # Note this submodule has now been initialized.
    _is_initted = True

    # Log this initialization attempt.
    logs.log_debug('Registering custom matplotlib colormaps...')

    # ..................{ COLOURS                            }..................
    # Primary colours.
    BLACK = (  0,   0,   0)
    GREEN = (  0, 255,   0)
    RED   = (255,   0,   0)
    BLUE  = (  0,   0, 255)

    # Secondary colours.
    CYAN    = (  9, 232, 239)
    MAGENTA = (239,  52, 236)
    # ORANGE  = (255, 164,  61)

    GREY_DARK = (51, 51, 51)

    GREEN_LIGHT  = (184, 255, 104)
    PURPLE_LIGHT = (219, 104, 255)

    BLUE_PALE = ( 56, 132, 255)
    RED_PALE  = (244,  66,  66)

    AQUA =  (53, 255, 211)
    AQUA2 = (71, 255, 218)

    GOLD   = (255, 231,  55)
    YELLOW = (255, 246, 130)

    # Salmon orange and/or red.
    SALMON  = (255, 111, 54)
    SALMON2 = (255, 117, 71)

    # ..................{ COLORMAPS                          }..................
    # Tuple of all application-specific colormaps iteratively registered below.
    COLORMAP_SCHEMES = (
        # Black-based colormaps.
        _MplColormapScheme(name='betse_electric_cyan',    colors=(BLACK, CYAN)),
        _MplColormapScheme(name='betse_electric_gold',    colors=(BLACK, YELLOW)),
        _MplColormapScheme(name='betse_electric_green',   colors=(BLACK, GREEN)),
        _MplColormapScheme(name='betse_electric_magenta', colors=(BLACK, MAGENTA)),
        _MplColormapScheme(name='betse_electric_orange',  colors=(BLACK, SALMON2)),
        _MplColormapScheme(name='betse_electric_blue',    colors=(BLACK, BLUE_PALE)),

        # Grey-based colormaps.
        _MplColormapScheme(name='betse_blue_chalkboard',    colors=(GREY_DARK, BLUE_PALE)),
        _MplColormapScheme(name='betse_cyan_chalkboard',    colors=(GREY_DARK, CYAN)),
        _MplColormapScheme(name='betse_gold_chalkboard',    colors=(GREY_DARK, GOLD)),
        _MplColormapScheme(name='betse_green_chalkboard',   colors=(GREY_DARK, GREEN)),
        _MplColormapScheme(name='betse_magenta_chalkboard', colors=(GREY_DARK, MAGENTA)),
        _MplColormapScheme(name='betse_orange_chalkboard',  colors=(GREY_DARK, SALMON)),

        # Salmon-based colormaps.
        _MplColormapScheme(name='betse_alien_chalkboard', colors=(SALMON2, GREY_DARK, AQUA2)),
        _MplColormapScheme(name='betse_alien_pale',       colors=(SALMON2, BLACK, AQUA2)),
        _MplColormapScheme(name='betse_alien_solid',      colors=(SALMON, BLACK, AQUA)),

        # Purple-based colormaps.
        _MplColormapScheme(name='betse_purple_green_chalkboard', colors=(MAGENTA, GREY_DARK, GREEN)),
        _MplColormapScheme(name='betse_purple_green_pale',       colors=(PURPLE_LIGHT, BLACK, GREEN_LIGHT)),
        _MplColormapScheme(name='betse_purple_green_solid',      colors=(MAGENTA, BLACK, GREEN)),

        # Blue-based colormaps.
        _MplColormapScheme(name='betse_red_blue_chalkboard', colors=(BLUE, GREY_DARK, RED)),
        _MplColormapScheme(name='betse_red_blue_pale',       colors=(BLUE_PALE, BLACK, RED_PALE)),
        _MplColormapScheme(name='betse_red_blue_solid',      colors=(BLUE, BLACK, RED)),
    )

    # ..................{ REGISTRATION                       }..................
    # Register each such colormap with matplotlib.
    for colormap_scheme in COLORMAP_SCHEMES:
        colormap_scheme.register()

# ....................{ PRIVATE ~ globals                  }....................
_is_initted = False
'''
``True`` only if the :func:`init` function has already been called under the
active Python interpreter.
'''

# ....................{ PRIVATE ~ classes                  }....................
class _MplColormapScheme(object):
    '''
    Matplotlib-specific **colormap scheme** (i.e., collection of parameters
    sufficient to subsequently define a standard linear-segmented colormap).

    Instances of this class are typically passed to the :func:`add_colormap`
    function, which both creates and registers a new colormap from the passed
    colormap scheme.

    Attributes
    -----------
    colormap_name : str
        Name of the colormap to be created.
    colors : SequenceTypes
        Two-dimensional sequence whose:

        * First dimension indexes each color defining this colormap's gradient.
          This dimension *must* be a sequence containing two or more colors.
        * Second dimension is a 3-sequence indexing the red, green, and blue
          (RGB) values defining this color.
    gamma : float
        Gamma curve value, adjusting the brightness of this colormap's
        **endpoint colors** (i.e., the colors at the bottom and top of this
        colormap). Specifically, matplotlib documentation internally states:

            colormap values are modified as c^gamma, where gamma is (1-beta)
            for beta>0 and 1/(1+beta) for beta<=0
    '''

    # ..................{ INITIALIZERS                      }..................
    def __init__(
        self,

        # Mandatory parameters.
        name: str,
        colors: SequenceTypes,

        # Optional parameters. All default values defined below should ideally
        # be identical to the same default values defined by matplotlib.
        gamma: float = 1.0,
    ) -> None:
        '''
        Initialize this colormap scheme.

        Parameters
        -----------
        name : str
            Name of the colormap to be created.
        colors : SequenceTypes
            Two-dimensional sequence whose:
            * First dimension indexes each color defining this colormap's
              gradient. This dimension *must* be a sequence containing two or
              more colors.
            * Second dimension is a 3-sequence indexing the red, green, and blue
              (RGB) values defining this color.
        gamma : float
            Gamma curve value, adjusting the brightness of this colormap's
            **endpoint colors** (i.e., the colors at the bottom and top of this
            colormap). Specifically, matplotlib documentation internally states:

                colormap values are modified as c^gamma, where gamma is (1-beta)
                for beta>0 and 1/(1+beta) for beta<=0

        Raises
        -----------
        BetseSequenceException
            If either:

            * The first dimension of ``colors`` contains less than two colors
              (i.e., is either empty *or* contains only one color).
            * Any second dimension of ``colors`` does *not* contain exactly
              three color values.
        BetseIntException
            If any color value in any second dimension of ``colors`` is *not*
            a valid color value in the range ``[0, 255]``.
        '''

        #FIXME: Raise an exception if this colormap name conflicts with that
        #of an existing colormap -- probably by defining a new
        #die_if_colormap() utility function in this submodule.

        # If this colormap defines less than two colors, raise an exception.
        if len(colors) < 2:
            #FIXME: Raise a less ambiguous "BetseMplColormapException" instead.
            raise BetseSequenceException(
                'Colormap scheme defines less than two colors: {!r}'.format(
                    colors))

        # For each color defining this colormap...
        for color in colors:
            # If this color is *NOT* a 3-sequence, raise an exception.
            sequences.die_unless_length(sequence=color, length=3)

            # If any of this color's values is *NOT* a valid RGB colour (i.e.,
            # integer in the range [0, 255]), raise an exception.
            ints.die_unless_byte(*color)

        # Classify all passed parameters.
        self._name = name
        self._colors = colors
        self._gamma = gamma

    # ..................{ REGISTERERS                        }..................
    @type_check
    def register(self) -> Colormap:
        '''
        Create a linear-segmented colormap from the current scheme, register
        this colormap with matplotlib, and return this colormap.
        '''

        # Log this registration attempt.
        # logs.log_debug(
        #     'Registering colormap "%s": %s',
        #     self.colormap_name, self.colors_normalized)

        # Two-dimensional Numpy array, normalizing each of each color's values
        # from [0, 255] to [0.0, 1.0] (while preserving the order of colors).
        colors_normalized = np.array(self._colors) / 255

        # Colormap synthesized from this colormap name and colors.
        #
        # Unfortunately, as the names of the first two parameters accepted by
        # this function have changed across matplotlib versions, these
        # parameters *MUST* be passed positionally for safety.
        colormap = LinearSegmentedColormap.from_list(
            self._name, colors_normalized, N=256, gamma=self._gamma)

        # Register this colormap with matplotlib.
        colormaps.register(colormap)

        # Return this colormap.
        return colormap




#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **class decorator** (i.e., class dynamically wrapping other classes
and callables at runtime) facilities.
'''

# ....................{ IMPORTS                           }....................
from abc import ABCMeta, abstractmethod
from betse.util.type.types import (
    type_check, CallableTypes, ClassType, MethodType)
from betse.util.type.decorator.decorators import decorator_chain

# ....................{ DECORATORS ~ method               }....................
class MethodDecoratorABC(object, metaclass=ABCMeta):
    '''
    Abstract base class of all **method decorators** (i.e., decorators *only*
    decorating methods bound to class instances), implemented as a class
    descriptor satisfying the standard descriptor protocol.

    This superclass efficiently caches bound methods on behalf of subclasses,
    guaranteeing all subclasses to be efficiently callable as proper methods.

    Attributes
    ----------
    _method_unbound: CallableTypes
        Unbound method (i.e., function) to be decorated.
    _obj_id_to_method_bound : dict
        Dictionary mapping from the unique identifier associated with each
        class instance containing a method decorated by this subclass to the
        same method bound to that class instance. While technically optional,
        the cache implemented by this dictionary avoids the need to recreate
        bound methods on each call to the :meth:`__get__` method.
    '''

    # ..................{ INITIALIZERS                      }..................
    @type_check
    def __init__(self, method: CallableTypes) -> None:
        '''
        Initialize this method decorator.

        Parameters
        ----------
        method: CallableTypes
            Unbound method (i.e., function) to be decorated.
        '''

        # Classify all passed parameters.
        self._method_unbound = method

        # Initialize all remaining instance variables.
        self._obj_id_to_method_bound = {}

    # ..................{ DESCRIPTORS                       }..................
    def __get__(self, obj: object, cls: ClassType) -> MethodType:
        '''
        Create, cache, and return a decorated method bound to the passed
        object.

        This method satisfies the descriptor protocol in a similar manner to
        Python itself. Python implicitly converts each function in a class body
        into a descriptor implementing the ``__get__()`` special method by
        internally creating and returning a copy of that function bound to the
        passed class instance.
        '''

        # If this descriptor is accessed as a class rather than instance
        # variable, return this low-level descriptor rather than the high-level
        # value to which this descriptor evaluates.
        if obj is None:
            return self

        # Unique identifier associated with this object.
        obj_id = id(obj)

        # Attempt to access the previously bound decorated method.
        try:
            self._obj_id_to_method_bound[obj_id]
        # If this method has yet to be bound, create and cache this bound
        # method. While superficially trivial, doing so is surprisingly
        # non-trivial when examined. In the following assignment:
        #
        # * Accessing "self.__call__" implicitly creates a new method from the
        #   __call__() function bound to the current instance of this class
        #   with the signature:
        #       def __call__bound(obj, *args, **kwargs)
        # * Instantiating "MethodType" implicitly creates a new method from the
        #   __call__bound() method bound to the passed instance of the parent
        #   class with the signature:
        #       def __call__bound_bound(*args, **kwargs)
        #
        # Hence, this bound method is actually a bound bound method (i.e., a
        # function bound to two class instances).
        except KeyError:
            self._obj_id_to_method_bound[obj_id] = MethodType(
                self.__call__, obj)

        # Return this bound method.
        return self._obj_id_to_method_bound[obj_id]

    # ..................{ CALLERS                           }..................
    def __call__(self, obj, *args, **kwargs) -> object:
        '''
        Call the decorated method previously passed to the :meth:`__init__`
        method bound to the passed object with the passed positional and
        keyword arguments, returning the value returned by this call.

        This special method is typically overriden by subclass implementations
        wrapping the decorated method with additional functionality.
        '''

        return self._method_unbound(obj, *args, **kwargs)

# ....................{ DECORATORS ~ property             }....................
abstractclassmethod = decorator_chain(classmethod, abstractmethod)
'''
Decorator decorating the passed superclass method as an **abstract class
method** (i.e., method called on a class rather than an instance of a class,
required to be defined by subclasses).

Motivation
----------
Technically, there already exist two standard means of decorating methods as
abstract properties:

* The deprecated :func:`abc.abstractproperty` decorator, which this decorator
  is intended to act as a drop-in replacement of. Since Python 3.3 unwisely
  deprecated that decorator, that decorator is *not* safely usable in
  perpetuity.
* Manually chaining the :class:`property` and :func:`abc.abstractmethod`
  decorators, which this decorator is intended to automate and hence replace.
  For unknown reasons, the Python community now treats manual chaining as a
  sensible replacement for the deprecated :func:`abc.abstractproperty`
  decorator. We disagree. Ergo, this decorator.
'''


abstractproperty = decorator_chain(property, abstractmethod)
'''
Decorator decorating the passed superclass method as an **abstract property**
(i.e., method implicitly called by Python on attempts to access instance
variables of the same, required to be defined by subclasses).

Motivation
----------
Technically, there already exist two standard means of decorating methods as
abstract properties:

* The deprecated :func:`abc.abstractproperty` decorator, which this decorator
  is intended to act as a drop-in replacement of. Since Python 3.3 unwisely
  deprecated that decorator, that decorator is *not* safely usable in
  perpetuity.
* Manually chaining the :class:`property` and :func:`abc.abstractmethod`
  decorators, which this decorator is intended to automate and hence replace.
  For unknown reasons, the Python community now treats manual chaining as a
  sensible replacement for the deprecated :func:`abc.abstractproperty`
  decorator. We disagree. Ergo, this decorator.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
**Mid-simulation animation** (i.e., animation produced *while* rather than
*after* solving a simulation) subclasses.
'''

# ....................{ IMPORTS                           }....................
import matplotlib
import numpy as np
from betse.lib.matplotlib.matplotlibs import mpl_config
from betse.science.math import mathunit
from betse.science.phase.phasecls import SimPhase
from betse.science.visual.anim.animabc import AnimCellsABC
from betse.util.type.types import type_check, SequenceTypes
from matplotlib import pyplot

# ....................{ SUBCLASSES                        }....................
#FIXME: Rename "_cell_data_plot" to "_cell_body_plot".
#FIXME: Rename "_cell_edges_plot" to "_cell_edge_plot".

class AnimCellsWhileSolving(AnimCellsABC):
    '''
    Context manager animating a mid-simulation animation.

    This manager animates arbitrary cell data as a time series plotted over the
    cell cluster (e.g., cell membrane voltage as a function of time) *while*
    rather than *after* solving a simulation phase.

    Caveats
    ----------
    This animation *must* be the target clause of a ``with`` statement,
    ensuring safe setup and teardown of this animation's non-blocking behavior.
    Likewise, this animation's :meth:`plot_frame` method *must* be called
    within the body of this ``with`` statement after producing for each sampled
    time step the cell data visualized by this animation for the corresponding
    frame.

    Attributes
    ----------
    _cell_edges_plot : LineCollection
        Plot of the current or prior frame's cell edges.
    _cell_data_plot : Collection
        Plot of the current or prior frame's cell contents.
    _cell_verts_id : int
        Unique identifier for the array of cell vertices (i.e.,
        `cells.cell_verts`) when plotting the current or prior frame.
        Retaining this identifier permits the `_plot_frame_figure()` method to
        efficiently detect and respond to physical changes (e.g., deformation
        forces, cutting events) in the fundamental structure of the previously
        plotted cell cluster.
    _is_colorbar_autoscaling_telescoped : bool
        ``True`` if colorbar autoscaling is permitted to increase but not
        decrease the colorbar range *or* ``False`` otherwise (i.e., if
        colorbar autoscaling is permitted to both increase and decrease the
        colorbar range). Such telescoping assists in emphasizing stable
        long-term patterns in cell data at a cost of deemphasizing unstable
        short-term patterns. If colorbar autoscaling is disabled (i.e.,
        ``is_color_autoscaled`` is ``False``), this will be ignored.
    '''

    # ..................{ INITIALIZERS                      }..................
    @type_check
    def __init__(
        self,

        # Mandatory parameters.
        phase: SimPhase,

        # Optional parameters.

        #FIXME: Permit this option to be configured via a new configuration
        #option in the YAML file.
        #FIXME: Actually, just excise this everywhere. Ally dislikes this
        #effect and I certainly never require it; ergo, remove.

        is_colorbar_autoscaling_telescoped: bool = False,
        *args, **kwargs
    ) -> None:
        '''
        Initialize this in-simulation animation.

        Parameters
        ----------
        phase: SimPhase
            Current simulation phase.
        is_colorbar_autoscaling_telescoped : optional[bool]
            ``True`` only if colorbar autoscaling is permitted to increase but
            not decrease the colorbar range *or* ``False`` otherwise (i.e., if
            colorbar autoscaling is permitted to both increase and decrease the
            colorbar range). Such telescoping assists in emphasizing stable
            long-term patterns in cell data at a cost of deemphasizing unstable
            short-term patterns. If colorbar autoscaling is disabled (i.e.,
            ``is_color_autoscaled`` is ``False``), this will be ignored.
            Defaults to ``False``.

        See the superclass method for all remaining parameters.
        '''

        # Initialize the superclass.
        super().__init__(
            *args,

            # Pass this simulation phase as is to our superclass.
            phase=phase,

            # Prevent the superclass from overlaying electric current or
            # concentration flux. Although this class does *NOT* animate a
            # streamplot, the time series required to plot this overlay is
            # unavailable until after the simulation ends.
            is_current_overlayable=False,

            # Save and show this mid-simulation animation only if this
            # configuration has enabled doing so.
            is_save=phase.p.anim.is_while_sim_save,
            is_show=phase.p.anim.is_while_sim_show,

            # Save this mid-simulation animation to a different parent
            # directory than that to which the corresponding post-simulation
            # animation is saved.
            save_dir_parent_basename='anim_while_solving',

            # Pass all remaining arguments as is to our superclass.
            **kwargs
        )
        # logs.log_debug('Showing mid-simulation animation: {}'.format(self._is_show))

        # Classify all remaining parameters.
        self._is_colorbar_autoscaling_telescoped = (
            is_colorbar_autoscaling_telescoped)

        # Unique identifier for the array of cell vertices. (See docstring.)
        self._cell_verts_id = id(self._phase.cells.cell_verts)

        #FIXME: This is a temp change until we get this right.
        #FIXME: Refactor to call the new
        #Cells.map_membranes_midpoint_to_cells_centre() method instead.

        # Vmem averaged over cell centres.
        vm_o = np.dot(self._phase.cells.M_sum_mems, self._phase.sim.vm) / (
            self._phase.cells.num_mems)

        # self._cell_time_series = self.sim.vm_time
        self._cell_time_series = self._phase.sim.vm_ave_time

        # cell_data_current = self.sim.vm
        cell_data_current = vm_o

        # Upscaled cell data for the first frame.
        cell_data = mathunit.upscale_units_milli(cell_data_current)

        # Collection of cell polygons with animated voltage data.
        #
        # If *NOT* simulating extracellular spaces, only animate intracellular
        # spaces.
        self._cell_data_plot = self._plot_cells_sans_ecm(
            cell_data=cell_data)

        # Perform all superclass plotting preparation immediately *BEFORE*
        # plotting this animation's first frame.
        self._prep_figure(
            color_mappables=self._cell_data_plot,
            color_data=cell_data,
        )

    # ..................{ CONTEXTS                          }..................
    def __enter__(self) -> 'AnimCellsWhileSolving':
        '''
        Enter the runtime context for this context manager, returning the same
        context manager as the value bound to the identifier in the ``as``
        clause of the ``with`` block using this context manager.

        This special method temporarily enables non-blocking matplotlib
        behavior for the duration of this plot or animation.
        '''

        # If displaying this animation, do so in a non-blocking manner.
        if self._is_show:
            # If the current matplotlib backend supports "true" non-blocking
            # behavior, prefer this non-deprecated approach.
            if mpl_config.is_backend_current_nonblockable():
                pyplot.show(block=False)
            # Else, fallback to the deprecated approach guaranteed to apply to
            # all matplotlib backends.
            else:
                matplotlib.interactive(True)
                pyplot.show()

        # Bind this animation to the "as" clause of this "with" block.
        return self


    def __exit__(self, exc_type, exc_val, exc_tb) -> bool:
        '''
        Exit the runtime context for this context manager given the passed
        exception metadata for an exception raised by the body of the ``with``
        block using this context manager if any, returning ``True`` only if
        that exception should be suppressed by this ``with`` block.

        This special method (in order):

        . Disables the non-blocking matplotlib behavior temporarily enabled by
          the prior :meth:`__enter__` call.
        . Safely closes this plot or animation.
        '''

        # Id displaying this animation *AND* the current matplotlib backend
        # fails to support "true" non-blocking behavior, disable the "fake"
        # non-blocking behavior enabled by the prior __enter__() call.
        if self._is_show and not mpl_config.is_backend_current_nonblockable():
            matplotlib.interactive(False)

        # Close this plot or animation.
        self.close()

        # Avoid suppressing exceptions raised by this "with" block.
        return False

    # ..................{ PLOTTERS                          }..................
    def _plot_frame_figure(self) -> None:

        # Upscaled cell data for the current time step.
        cell_data = mathunit.upscale_units_milli(
            self._cell_time_series[self._time_step])

        #FIXME: Duplicated from above. What we probably want to do is define a
        #new _get_cell_data() method returning this array in a centralized
        #manner callable both here and above. Rays of deluded beaming sunspray!

        # If the unique identifier for the array of cell vertices has *NOT*
        # changed, the cell cluster has *NOT* fundamentally changed and need
        # only be updated with this time step's cell data.
        if self._cell_verts_id == id(self._phase.cells.cell_verts):
            # loggers.log_info(
            #     'Updating animation "{}" cell plots...'.format(self._type))
            self._update_cell_plots(cell_data)
        # Else, the cell cluster has fundamentally changed (e.g., due to
        # physical deformations or cutting events) and must be recreated.
        else:
            # loggers.log_info(
            #     'Reviving animation "{}" cell plots...'.format(self._type))

            # Prevent subsequent calls to this method from erroneously
            # recreating the cell cluster again.
            self._cell_verts_id = id(self._phase.cells.cell_verts)

            # Recreate the cell cluster.
            self._revive_cell_plots(cell_data)

        # Update the color bar with the content of the cell body plot *AFTER*
        # possibly recreating this plot above.
        if self._conf.is_color_autoscaled:
            cell_data_vm = cell_data

            # If autoscaling this colorbar in a telescoping manner and this is
            # *NOT* the first time step, do so.
            #
            # If this is the first time step, the previously calculated minimum
            # and maximum colors are garbage and thus *MUST* be ignored.
            if (self._is_colorbar_autoscaling_telescoped and
                not self._is_time_step_first):
                self._color_min = min(self._color_min, np.min(cell_data_vm))
                self._color_max = max(self._color_max, np.max(cell_data_vm))
            # Else, autoscale this colorbar in an unrestricted manner.
            else:
                self._color_min = np.min(cell_data_vm)
                self._color_max = np.max(cell_data_vm)

            # Autoscale the colorbar to these colors.
            self._rescale_color_mappables()

        # If displaying this frame...
        if self._is_show:
            # If the current event loop is idle, draw this frame; else, noop.
            # As detailed by the pyplot.draw() docstring, this is the
            # object-oriented equivalent to calling pyplot.draw().
            self._figure.canvas.draw_idle()
            # self._figure.canvas.draw()


    @type_check
    def _update_cell_plots(self, cell_data: SequenceTypes) -> None:
        '''
        Update *without* recreating all cell plots for this time step with the
        passed array of arbitrary cell data.

        This method is intended to be called *unless* physical changes
        (e.g., deformation forces, cutting events) in the underlying structure
        of the cell cluster have occurred for this simulation time step.

        Parameters
        -----------
        cell_data : SequenceTypes
            Arbitrary cell data defined on an environmental grid to be plotted.
        '''

        self._update_cell_plot_sans_ecm(
            cell_plot=self._cell_data_plot,
            cell_data=cell_data)


    @type_check
    def _revive_cell_plots(self, cell_data: SequenceTypes) -> None:
        '''
        Recreate all cell plots for this time step with the passed array of
        arbitrary cell data.

        This method is intended to be called in response to physical changes
        (e.g., deformation forces, cutting events) in the underlying structure
        of the cell cluster for this simulation time step. This method is both
        inefficient and destructive, and should be called only when needed.

        Parameters
        -----------
        cell_data : SequenceTypes
            Arbitrary cell data defined on an environmental grid to be plotted.
        '''

        self._cell_data_plot = self._revive_cell_plots_sans_ecm(
            cell_plot=self._cell_data_plot,
            cell_data=cell_data)
